"""
===========================================
Comparison (:mod:`CADETProcess.comparison`)
===========================================

.. currentmodule:: CADETProcess.comparison

Classes for comparing simulation results with a reference.

Difference Metrics
==================

.. autosummary::
    :toctree: generated/

    DifferenceBase
    SSE
    RMSE
    NRMSE
    Norm
    L1
    L2
    AbsoluteArea
    RelativeArea
    Shape
    PeakHeight
    PeakPosition
    BreakthroughHeight
    BreakthroughPosition

Comparator
==========

.. autosummary::
    :toctree: generated/

    Comparator

"""

from .difference import *
from .comparator import *



"""
===================================================
Data Structures (:mod:`CADETProcess.dataStructure`)
===================================================

.. currentmodule:: CADETProcess.dataStructure

This module provides datastructures to simplify defining setters and getters for
(model) parameters.
It is mostly based on the data model introduced in [1]_

Notes
-----

At some point it might be considered to switch to attrs
(see `#15 <https://github.com/fau-advanced-separations/CADET-Process/issues/15>`_).

References
----------
.. [1] Jones, B. K., Beazley, D. (2013).
    Python Cookbook: Recipes for Mastering Python 3. United States: O'Reilly Media.


.. autosummary::
    :toctree: generated/

    dataStructure
    aggregator
    parameter
    parameter_group
    cache
    diskcache
    nested_dict

"""

from .dataStructure import *
from .aggregator import *
from .parameter import *
from .parameter_group import *
from .cache import *
from .diskcache import *
from .nested_dict import *
from .deprecation import *



import shutil
import unittest
from pathlib import Path

import numpy as np
from addict import Dict
from CADETProcess.optimization import U_NSGA3, OptimizationResults

from tests.test_optimization_problem import setup_optimization_problem
from tests.test_population import setup_population


class OptimizationResultsWithoutNans(OptimizationResults):
    """
    Patch class that removes None values in the OptimizationResults during .to_dict() call.
    This allows serialization with any version of CADET-Python pending the merge of
    https://github.com/cadet/CADET-Python/pull/48

    """

    def to_dict(self) -> dict:
        """Convert Results to a dictionary.

        Returns
        -------
        addict.Dict
            Results as a dictionary with populations stored as list of dictionaries.
        """
        data = Dict()
        data.system_information = self.system_information
        data.optimizer_state = self.optimizer_state
        data.population_all_id = str(self.population_all.id)
        data.populations = {i: pop.to_dict() for i, pop in enumerate(self.populations)}
        data.pareto_fronts = {
            i: front.to_dict() for i, front in enumerate(self.pareto_fronts)
        }
        if self._meta_fronts is not None:
            data.meta_fronts = {
                i: front.to_dict() for i, front in enumerate(self.meta_fronts)
            }
        if self.time_elapsed is not None:
            data.time_elapsed = self.time_elapsed
            data.cpu_time = self.cpu_time

        data = self._remove_none_values(data)

        return data

    @staticmethod
    def _remove_none_values(dictionary) -> dict:
        """Remove None values from a dictionary.

        Parameters
        ----------
        dictionary : dict
        Dictionary to remove None values from.

        Returns
        -------
        dict
        Dictionary with None values removed.
        """
        delete_keys = []
        for key, value in dictionary.items():
            if value is None:
                delete_keys.append(key)
            elif isinstance(value, dict):
                OptimizationResultsWithoutNans._remove_none_values(value)

        for key in delete_keys:
            del dictionary[key]

        return dictionary


def setup_optimization_problem_and_results(
    n_gen=3,
    n_ind=3,
    n_vars=2,
    n_obj=1,
    n_nonlin=0,
    n_meta=0,
    rng=None,
    initialize_data=True,
):
    optimization_problem = setup_optimization_problem(n_vars, n_obj, n_nonlin, n_meta)
    optimizer = U_NSGA3()

    optimization_results = OptimizationResultsWithoutNans(
        optimization_problem, optimizer
    )
    results_dir = Path("tmp") / "optimization_results"

    shutil.rmtree(results_dir, ignore_errors=True)
    results_dir.mkdir(exist_ok=True, parents=True)
    (results_dir / "figures").mkdir(exist_ok=True, parents=True)
    optimization_results.results_directory = results_dir

    if initialize_data:
        if rng is None:
            rng = np.random.default_rng(12345)

        for gen in range(n_gen):
            pop = setup_population(n_ind, n_vars, n_obj, n_nonlin, n_meta, rng)
            optimization_results.update(pop)
            optimization_results.update_pareto()
            if n_meta > 0:
                optimization_results.update_meta()

    return optimization_problem, optimization_results


class TestOptimizationResults(unittest.TestCase):
    def setUp(self):
        _, self.optimization_results = setup_optimization_problem_and_results()

    def test_history(self):
        n_gen_expected = 3
        n_gen = self.optimization_results.n_gen
        self.assertEqual(n_gen, n_gen_expected)

        n_evals_expected = 9
        n_evals = self.optimization_results.n_evals
        self.assertEqual(n_evals, n_evals_expected)

        f_min_history_expected = np.array(
            [
                [-0.79736546],
                [-0.94888115],
                [-0.94888115],
            ]
        )
        f_min_history = self.optimization_results.f_min_history
        np.testing.assert_almost_equal(f_min_history, f_min_history_expected)

        g_min_history_expected = None
        g_min_history = self.optimization_results.g_min_history
        self.assertIsNone(g_min_history, g_min_history_expected)

        m_min_history_expected = None
        m_min_history = self.optimization_results.m_min_history
        self.assertIsNone(m_min_history, m_min_history_expected)

    def test_serialization(self):
        data = self.optimization_results.to_dict()

        _, results_new = setup_optimization_problem_and_results(initialize_data=False)
        results_new.update_from_dict(data)
        data_new = results_new.to_dict()
        np.testing.assert_equal(data, data_new)

    def test_io_serialization(self):
        optimization_problem, _ = setup_optimization_problem_and_results()

        # save_results() needs to happen after creation of optimization_problem,
        # because during optimization_problem creation the results folder gets cleared
        # and that would delete the save
        self.optimization_results.save_results("checkpoint")

        optimizer = U_NSGA3()
        optimization_results_new = optimizer.load_results(
            checkpoint_path=self.optimization_results.results_directory
            / "checkpoint.h5",
            optimization_problem=optimization_problem,
        )
        np.testing.assert_equal(
            desired=self.optimization_results.to_dict(),
            actual=OptimizationResultsWithoutNans._remove_none_values(
                optimization_results_new.to_dict()
            ),
        )


if __name__ == "__main__":
    unittest.main()



import numpy as np
import pytest
from CADETProcess.numerics import round_to_significant_digits


def test_basic_functionality():
    values = np.array([123.456, 0.001234, 98765])
    expected = np.array([123.0, 0.00123, 98800.0])
    result = round_to_significant_digits(values, 3)
    np.testing.assert_allclose(result, expected, rtol=1e-6)


def test_handling_zeros():
    values = np.array([0, 0.0, -0.0])
    expected = np.array([0.0, 0.0, -0.0])
    result = round_to_significant_digits(values, 3)
    np.testing.assert_array_equal(result, expected)


def test_negative_numbers():
    values = np.array([-123.456, -0.001234, -98765])
    expected = np.array([-123.0, -0.00123, -98800.0])
    result = round_to_significant_digits(values, 3)
    np.testing.assert_allclose(result, expected, rtol=1e-6)


def test_large_numbers():
    values = np.array([1.23e10, 9.87e15])
    expected = np.array([1.23e10, 9.87e15])
    result = round_to_significant_digits(values, 3)
    np.testing.assert_allclose(result, expected, rtol=1e-6)


def test_small_numbers():
    values = np.array([1.23e-10, 9.87e-15])
    expected = np.array([1.23e-10, 9.87e-15])
    result = round_to_significant_digits(values, 3)
    np.testing.assert_allclose(result, expected, rtol=1e-6)


def test_mixed_values():
    values = np.array([123.456, 0, -0.001234, 9.8765e-5])
    expected = np.array([123.0, 0.0, -0.00123, 9.88e-5])
    result = round_to_significant_digits(values, 3)
    np.testing.assert_allclose(result, expected, rtol=1e-6)


def test_invalid_digits():
    with pytest.raises(
        ValueError, match="Number of significant digits must be greater than 0."
    ):
        round_to_significant_digits(np.array([123.456]), 0)


def test_empty_array():
    values = np.array([])
    expected = np.array([])
    result = round_to_significant_digits(values, 3)
    np.testing.assert_array_equal(result, expected)


def test_non_array_input():
    values = [123.456, 0.001234, 98765]  # List input
    expected = np.array([123.0, 0.00123, 98800.0])
    result = round_to_significant_digits(values, 3)
    np.testing.assert_allclose(result, expected, rtol=1e-6)


def test_none_digits():
    values = [123.456, 0.001234, 98765]  # List input
    expected = np.array([123.456, 0.001234, 98765])
    result = round_to_significant_digits(values, None)
    np.testing.assert_allclose(result, expected, rtol=1e-6)


def test_nan_digits():
    values = np.array([123.456, np.nan, 98765])
    expected = np.array([123.0, np.nan, 98800.0])
    result = round_to_significant_digits(values, 3)
    np.testing.assert_allclose(result, expected, rtol=1e-6)



# ---
# jupyter:
#   jupytext:
#     formats: md:myst,py:percent
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.17.1
#   kernelspec:
#     display_name: Python 3
#     name: python3
# ---

# %% [markdown]
# (batch_elution_example)=
# # Batch Elution Chromatography
# tags: batch elution, langmuir
#
# A basic chromatographic batch-elution setup is comprised of feed and eluent reservoirs, a pump that can deliver the necessary flow rate against the pressure drop of the packed column, a valve to select whether feed or eluent are pumped into the column, the column itself, and one or more valves for the collection of fractions.
#
# In **CADET-Process**, this can be modelled by connecting two {class}`Inlets <CADETProcess.processModel.Inlet>`, a column model (e.g. {class}`~CADETProcess.processModel.LumpedRateModelWithoutPores`), and an {class}`~CADETProcess.processModel.Outlet`.
#
# ```{figure} ./figures/flow_sheet.svg
# Flow sheet for batch elution process.
# ```
#
# For this example, consider a two-component system with a Langmuir isotherm.
#
# ## Component System

# %%
from CADETProcess.processModel import ComponentSystem

component_system = ComponentSystem()
component_system.add_component('A')
component_system.add_component('B')

# %% [markdown]
# ## Binding Model

# %%
from CADETProcess.processModel import Langmuir

binding_model = Langmuir(component_system, name='langmuir')
binding_model.is_kinetic = False
binding_model.adsorption_rate = [0.02, 0.03]
binding_model.desorption_rate = [1, 1]
binding_model.capacity = [100, 100]

# %% [markdown]
# ## Unit Operations

# %%
from CADETProcess.processModel import (
    Inlet, LumpedRateModelWithoutPores, Outlet
)
feed = Inlet(component_system, name='feed')
feed.c = [10, 10]

eluent = Inlet(component_system, name='eluent')
eluent.c = [0, 0]

column = LumpedRateModelWithoutPores(component_system, name='column')
column.binding_model = binding_model

column.length = 0.6
column.diameter = 0.024
column.axial_dispersion = 4.7e-7
column.total_porosity = 0.7

column.solution_recorder.write_solution_bulk = True

outlet = Outlet(component_system, name='outlet')

# %% [markdown]
# ## Flow Sheet

# %%
from CADETProcess.processModel import FlowSheet

flow_sheet = FlowSheet(component_system)

flow_sheet.add_unit(feed, feed_inlet=True)
flow_sheet.add_unit(eluent, eluent_inlet=True)
flow_sheet.add_unit(column)
flow_sheet.add_unit(outlet, product_outlet=True)

flow_sheet.add_connection(feed, column)
flow_sheet.add_connection(eluent, column)
flow_sheet.add_connection(column, outlet)

# %% [markdown]
# ## Process
# To model the injection valve, {class}`Events <CADETProcess.processModel.Event>` are introduced that modify the {attr}`~CADETProcess.procesModel.Inlet.flow_rate` attribute of the {class}`~CADETProcess.processModel.Inlet` unit operations.
#
# ```{figure} ./figures/events.svg
# Events of batch elution process.
# ```

# %%
from CADETProcess.processModel import Process
process = Process(flow_sheet, 'batch elution')

Q = 60 / (60 * 1e6)
process.add_event('feed_on', 'flow_sheet.feed.flow_rate', Q)
process.add_event('feed_off', 'flow_sheet.feed.flow_rate', 0.0)

process.add_event('eluent_on', 'flow_sheet.eluent.flow_rate', Q)
process.add_event('eluent_off', 'flow_sheet.eluent.flow_rate', 0.0)

# %% [markdown]
# ### Event Dependencies
# To reduce the number of event times that need to be specified, event dependencies are specified which enforce that always either feed or eluent are being pumped through the column.
#
# ```{figure} ./figures/event_dependencies.svg
# Events of batch elution process with event dependencies.
# ```

# %%
process.add_duration('feed_duration')

process.add_event_dependency('eluent_on', ['feed_off'])
process.add_event_dependency('eluent_off', ['feed_on'])
process.add_event_dependency('feed_off', ['feed_on', 'feed_duration'], [1, 1])

# %% [markdown]
# ### Event Times
# Now, the cycle time is set to $10~min$ and the `feed_duration` to $1~min$.

# %%
process.cycle_time = 600
process.feed_duration.time = 60

# %% [markdown]
# ## Simulate Process

# %%
if __name__ == '__main__':
    from CADETProcess.simulator import Cadet
    process_simulator = Cadet()

    simulation_results = process_simulator.simulate(process)

# %% [markdown]
# ## Optimize Fractionation Times
# After simulation, the {class}`~CADETProcess.simulationResults.SimulationResults` can be analyzed to determine optimal fractionation times using the {mod}`~CADETProcess.fractionation` module.

# %%
if __name__ == '__main__':
    from CADETProcess.fractionation import FractionationOptimizer
    fractionation_optimizer = FractionationOptimizer()

    fractionator = fractionation_optimizer.optimize_fractionation(
        simulation_results, purity_required=[0.95, 0.95]
    )

    print(fractionator.performance)
    _ = fractionator.plot_fraction_signal()

# %% [markdown]
# By selecting appropriate operating conditions, such as injected amount and flow rate, an efficient operating scenario can be achieved in which the stationary phase is utilized very efficiently.
# The highest product recovery is achieved by "baseline separation," where the component peaks of the same injection do not overlap when leaving the column.
# Moreover, by minimizing the time between two injections, productivity can be improved.
# By allowing for waste fractions collected between product fractions or between peaks of consecutive injections, productivity and eluent consumption may be further improved at the cost of lower recovery.
#
# These operating conditions can be adjusted using model based design.
# For this purpose, an {class}`~CADETProcess.optimization.OptimizationProblem` is set up where to maximize process performance.
# This can be achieved by combining multiple parameters into a single objective (see {ref}`batch_elution_optimization_single`) or by setting up a multi-objective problem (see {ref}`batch_elution_optimization_multi`).
#
#
# ```{toctree}
# :maxdepth: 1
# :hidden:
#
# optimization_single
# optimization_multi
# ```



from __future__ import annotations

import csv
import os
import warnings
from functools import wraps
from pathlib import Path
from typing import TYPE_CHECKING, Any, Literal, Optional

import matplotlib.cm as cmx
import matplotlib.colors as colors
import matplotlib.pyplot as plt
import numpy as np
from addict import Dict
from cadet import H5
from matplotlib.axes import Axes
from matplotlib.figure import Figure

from CADETProcess import CADETProcessError, plotting
from CADETProcess.dataStructure import (
    Bool,
    Dictionary,
    String,
    Structure,
    UnsignedFloat,
    UnsignedInteger,
)
from CADETProcess.optimization import Individual, ParetoFront, Population
from CADETProcess.sysinfo import system_information

if TYPE_CHECKING:
    from CADETProcess.optimization import OptimizationProblem, OptimizerBase

cmap_feas = plt.get_cmap("winter_r")
cmap_infeas = plt.get_cmap("autumn_r")


class OptimizationResults(Structure):
    """
    Optimization results.

    Attributes
    ----------
    optimization_problem : OptimizationProblem
        Optimization problem.
    optimizer : OptimizerBase
        Optimizer used to optimize the OptimizationProblem.
    success : bool
        True if optimization was successfully terminated. False otherwise.
    exit_flag : int
        Information about the solver termination.
    exit_message : str
        Additional information about the solver status.
    time_elapsed : float
        Execution time of simulation.
    cpu_time : float
        CPU run time, taking into account the number of cores used for the optimiation.
    system_information : dict
        Information about the system on which the optimization was performed.
    x : list
        Values of optimization variables at optimum.
    f : np.ndarray
        Value of objective function at x.
    g : np.ndarray
        Values of constraint function at x
    population_last : Population
        Last population.
    pareto_front : ParetoFront
        Pareto optimal solutions.
    meta_front : ParetoFront
        Reduced pareto optimal solutions using meta scores and multi-criteria decision
        functions.
    """

    success = Bool(default=False)
    exit_flag = UnsignedInteger()
    exit_message = String()
    time_elapsed = UnsignedFloat()
    cpu_time = UnsignedFloat()
    system_information = Dictionary()

    def __init__(
        self,
        optimization_problem: OptimizationProblem,
        optimizer: OptimizerBase,
        similarity_tol: float = 0,
    ) -> None:
        """Initialize OptimizationResults object."""
        self.optimization_problem = optimization_problem
        self.optimizer = optimizer

        self._optimizer_state = Dict()

        self._population_all = Population()
        self._populations = []
        self._similarity_tol = similarity_tol
        self._pareto_fronts = []

        if optimization_problem.n_multi_criteria_decision_functions > 0:
            self._meta_fronts = []
        else:
            self._meta_fronts = None

        self.results_directory = None

        self.system_information = system_information

    @property
    def results_directory(self) -> Path:
        """Path: Results directory path."""
        return self._results_directory

    @results_directory.setter
    def results_directory(self, results_directory: str | os.PathLike) -> None:
        if results_directory is not None:
            results_directory = Path(results_directory)
            self.plot_directory = Path(results_directory / "figures")
            self.plot_directory.mkdir(exist_ok=True)
        else:
            self.plot_directory = None

        self._results_directory = results_directory

    @property
    def is_finished(self) -> bool:
        """bool: True if optimization has finished. False otherwise."""
        if self.exit_flag is None:
            return False
        else:
            return True

    @property
    def optimizer_state(self) -> dict:
        """dict: Internal state of the optimizer."""
        return self._optimizer_state

    @property
    def populations(self) -> list[Population]:
        """list[Population]: List of populations per generation."""
        return self._populations

    @property
    def population_last(self) -> Population:
        """Population: Final population."""
        return self.populations[-1]

    @property
    def population_all(self) -> Population:
        """Population: Population with all evaluated individuals."""
        return self._population_all

    @property
    def pareto_fronts(self) -> list[ParetoFront]:
        """list[ParetoFront]: List of Pareto fronts per generation."""
        return self._pareto_fronts

    @property
    def pareto_front(self) -> ParetoFront:
        """ParetoFront: Final Pareto front."""
        return self._pareto_fronts[-1]

    @property
    def meta_fronts(self) -> list[Population]:
        """list[Population]: List of meta fronts per generation."""
        if self._meta_fronts is None:
            return self.pareto_fronts
        else:
            return self._meta_fronts

    @property
    def meta_front(self) -> Population:
        """Population: Meta front."""
        if self._meta_fronts is None:
            return self.pareto_front
        else:
            return self._meta_fronts[-1]

    def update(self, new: Individual | Population) -> None:
        """
        Update Results.

        Parameters
        ----------
        new : Individual, Population
            New results

        Raises
        ------
        CADETProcessError
            If new is not an instance of Individual or Population
        """
        if isinstance(new, Individual):
            population = Population()
            population.add_individual(new)
        elif isinstance(new, Population):
            population = new
        else:
            raise CADETProcessError("Expected Population or Individual")
        self._populations.append(population)
        self.population_all.update(population)

    def update_pareto(self, pareto_new: Population | None = None) -> None:
        """
        Update pareto front with new population.

        Parameters
        ----------
        pareto_new : Population, optional
            New pareto front. If None, update existing front with latest population.
        """
        pareto_front = ParetoFront(similarity_tol=self._similarity_tol)

        if pareto_new is not None:
            pareto_front.update_population(pareto_new)
        else:
            if len(self.pareto_fronts) > 0:
                pareto_front.update_population(self.pareto_front)
            pareto_front.update_population(self.population_last)

        if self._similarity_tol:
            pareto_front.remove_similar()
        self._pareto_fronts.append(pareto_front)

    def update_meta(self, meta_front: Population) -> None:
        """
        Update meta front with new population.

        Parameters
        ----------
        meta_front : Population
            New meta front.
        """
        if self._similarity_tol:
            meta_front.remove_similar()
        self._meta_fronts.append(meta_front)

    @property
    def n_evals(self) -> int:
        """int: Number of evaluations."""
        return sum([len(pop) for pop in self.populations])

    @property
    def n_gen(self) -> int:
        """int: Number of generations."""
        return len(self.populations)

    @property
    def x(self) -> np.ndarray:
        """np.ndarray: Optimal points in untransformed space."""
        return self.meta_front.x

    @property
    def x_transformed(self) -> np.ndarray:
        """np.ndarray: Optimal points in transformed space."""
        return self.meta_front.x_transformed

    @property
    def cv_bounds(self) -> np.ndarray:
        """np.ndarray: Bound constraint violation of optimal points."""
        return self.meta_front.cv_bounds

    @property
    def cv_lincon(self) -> np.ndarray:
        """np.ndarray: Linear constraint violation of optimal points."""
        return self.meta_front.cv_lincon

    @property
    def cv_lineqcon(self) -> np.ndarray:
        """np.ndarray: Linear equality constraint violation of optimal points."""
        return self.meta_front.cv_lineqcon

    @property
    def f(self) -> np.ndarray:
        """np.ndarray: Objective function values of optimal points."""
        return self.meta_front.f

    @property
    def g(self) -> np.ndarray:
        """np.ndarray: Nonlinear constraint function values of optimal points."""
        return self.meta_front.g

    @property
    def cv_nonlincon(self) -> np.ndarray:
        """np.ndarray: Nonlinear constraint violation values of optimal points."""
        return self.meta_front.cv_nonlincon

    @property
    def m(self) -> np.ndarray:
        """np.ndarray: Meta scores of optimal points."""
        return self.meta_front.m

    @property
    def n_evals_history(self) -> np.ndarray:
        """int: Number of evaluations per generation."""
        n_evals = [len(pop) for pop in self.populations]
        return np.cumsum(n_evals)

    @property
    def f_best_history(self) -> np.ndarray:
        """np.ndarray: Best objective values per generation."""
        return np.array([pop.f_best for pop in self.meta_fronts])

    @property
    def f_min_history(self) -> np.ndarray:
        """np.ndarray: Minimum objective values per generation."""
        return np.array([pop.f_min for pop in self.meta_fronts])

    @property
    def f_max_history(self) -> np.ndarray:
        """np.ndarray: Maximum objective values per generation."""
        return np.array([pop.f_max for pop in self.meta_fronts])

    @property
    def f_avg_history(self) -> np.ndarray:
        """np.ndarray: Average objective values per generation."""
        return np.array([pop.f_avg for pop in self.meta_fronts])

    @property
    def g_best_history(self) -> np.ndarray:
        """np.ndarray: Best nonlinear constraint per generation."""
        return np.array([pop.g_best for pop in self.meta_fronts])

    @property
    def g_min_history(self) -> np.ndarray:
        """np.ndarray: Minimum nonlinear constraint values per generation."""
        if self.optimization_problem.n_nonlinear_constraints == 0:
            return None
        else:
            return np.array([pop.g_min for pop in self.meta_fronts])

    @property
    def g_max_history(self) -> np.ndarray:
        """np.ndarray: Maximum nonlinear constraint values per generation."""
        if self.optimization_problem.n_nonlinear_constraints == 0:
            return None
        else:
            return np.array([pop.g_max for pop in self.meta_fronts])

    @property
    def g_avg_history(self) -> np.ndarray:
        """np.ndarray: Average nonlinear constraint values per generation."""
        if self.optimization_problem.n_nonlinear_constraints == 0:
            return None
        else:
            return np.array([pop.g_avg for pop in self.meta_fronts])

    @property
    def cv_nonlincon_min_history(self) -> np.ndarray:
        """np.ndarray: Minimum nonlinear constraint violation values per generation."""
        if self.optimization_problem.n_nonlinear_constraints == 0:
            return None
        else:
            return np.array([pop.cv_nonlincon_min for pop in self.meta_fronts])

    @property
    def cv_nonlincon_max_history(self) -> np.ndarray:
        """np.ndarray: Maximum nonlinear constraint violation values per generation."""
        if self.optimization_problem.n_nonlinear_constraints == 0:
            return None
        else:
            return np.array([pop.cv_nonlincon_max for pop in self.meta_fronts])

    @property
    def cv_nonlincon_avg_history(self) -> np.ndarray:
        """np.ndarray: Average nonlinear constraint violation values per generation."""
        if self.optimization_problem.n_nonlinear_constraints == 0:
            return None
        else:
            return np.array([pop.cv_nonlincon_avg for pop in self.meta_fronts])

    @property
    def m_best_history(self) -> np.ndarray:
        """np.ndarray: Best meta scores per generation."""
        return np.array([pop.m_best for pop in self.meta_fronts])

    @property
    def m_min_history(self) -> np.ndarray:
        """np.ndarray: Minimum meta scores per generation."""
        if self.optimization_problem.n_meta_scores == 0:
            return None
        else:
            return np.array([pop.m_min for pop in self.meta_fronts])

    @property
    def m_max_history(self) -> np.ndarray:
        """np.ndarray: Maximum meta scores per generation."""
        if self.optimization_problem.n_meta_scores == 0:
            return None
        else:
            return np.array([pop.m_max for pop in self.meta_fronts])

    @property
    def m_avg_history(self) -> np.ndarray:
        """np.ndarray: Average meta scores per generation."""
        if self.optimization_problem.n_meta_scores == 0:
            return None
        else:
            return np.array([pop.m_avg for pop in self.meta_fronts])

    def plot_figures(self, show: bool = True) -> None:
        """
        Plot result figures.

        See Also
        --------
        plot_convergence
        plot_objectives
        plot_corner
        plot_pairwise
        plot_pareto
        """
        if self.plot_directory is None:
            return

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")

            self.plot_convergence(
                "objectives", show=show, plot_directory=self.plot_directory
            )
            if self.optimization_problem.n_nonlinear_constraints > 0:
                self.plot_convergence(
                    "nonlinear_constraints",
                    show=show,
                    plot_directory=self.plot_directory,
                )
            if self.optimization_problem.n_meta_scores > 0:
                self.plot_convergence(
                    "meta_scores", show=show, plot_directory=self.plot_directory
                )
            self.plot_objectives(show=show, plot_directory=self.plot_directory)
            if self.optimization_problem.n_variables > 1 and len(self.x) > 1:
                self.plot_corner(show=show, plot_directory=self.plot_directory)

            self.plot_pairwise(show=show, plot_directory=self.plot_directory)

            if self.optimization_problem.n_objectives > 1:
                self.plot_pareto(
                    show=show,
                    plot_directory=self.plot_directory,
                    plot_evolution=True,
                    plot_pareto=False,
                )

    def plot_objectives(
        self,
        include_meta: bool = True,
        plot_pareto: bool = False,
        plot_infeasible: bool = True,
        plot_individual: bool = False,
        autoscale: bool = True,
        show: bool = True,
        plot_directory: Optional[str | Path] = None,
    ) -> None:
        """
        Plot objective function values for all optimization generations.

        Parameters
        ----------
        include_meta : bool, optional
            If True, meta scores will be included in the plot. The default is True.
        plot_pareto : bool, optional
            If True, only plot Pareto front members of each generation are plotted.
            Else, all evaluated individuals are plotted.
            The default is False.
        plot_infeasible : bool, optional
            If True, plot infeasible points. The default is True.
        plot_individual : bool, optional
            If True, create separate figures for each objective. Otherwise, all
            objectives are plotted in one figure.
            The default is False.
        plot_infeasible : bool, optional
            If True, plot infeasible points. The default is False.
        autoscale : bool, optional
            If True, automatically adjust the scaling of the axes. The default is True.
        show : bool, optional
            If True, display the plot. The default is True.
        plot_directory : str, optional
            The directory where the plot should be saved.
            The default is None.

        Returns
        -------
        tuple
            A tuple containing:
            - plt.Figure: The Matplotlib Figure object.
            - npt.NDArray[plt.Axes]: An array of Axes objects representing the subplots.

        See Also
        --------
        CADETProcess.optimization.Population.plot_objectives
        """
        axs = None
        figs = None
        _show = False
        _plot_directory = None

        cNorm = colors.Normalize(vmin=0, vmax=self.n_gen)
        scalarMap_feas = cmx.ScalarMappable(norm=cNorm, cmap=cmap_feas)
        scalarMap_infeas = cmx.ScalarMappable(norm=cNorm, cmap=cmap_infeas)

        if plot_pareto:
            populations = self.pareto_fronts
            population_last = self.pareto_front
        else:
            populations = self.populations
            population_last = self.population_last

        for i, gen in enumerate(populations):
            if gen is population_last:
                _plot_directory = plot_directory
                _show = show
            figs, axs = gen.plot_objectives(
                figs,
                axs,
                include_meta=include_meta,
                plot_infeasible=plot_infeasible,
                plot_individual=plot_individual,
                autoscale=autoscale,
                color_feas=scalarMap_feas.to_rgba(i),
                color_infeas=scalarMap_infeas.to_rgba(i),
                show=_show,
                plot_directory=_plot_directory,
            )

        return figs, axs

    def plot_pareto(
        self,
        show: bool = True,
        plot_pareto: bool = True,
        plot_evolution: bool = False,
        plot_directory: Optional[str | Path] = None,
    ) -> None:
        """
        Plot Pareto fronts for each generation in the optimization.

        The Pareto front represents the optimal solutions that cannot be improved in one
        objective without sacrificing another.
        The method shows a pairwise Pareto plot, where each objective is plotted against
        every other objective in a scatter plot, allowing for a visualization of the
        trade-offs between the objectives.
        To highlight the progress, a colormap is used where later generations are
        plotted with darker blueish colors.

        Parameters
        ----------
        show : bool, optional
            If True, display the plot.
            The default is True.
        plot_pareto : bool, optional
            If True, only Pareto front members of each generation are plotted.
            Else, all evaluated individuals are plotted.
            The default is True.
        plot_evolution : bool, optional
            If True, the Pareto front is plotted for each generation.
            Else, only final Pareto front is plotted.
            The default is False.
        plot_directory : str, optional
            The directory where the plot should be saved.
            The default is None.

        Returns
        -------
        tuple
            A tuple containing:
            - plt.Figure: The Matplotlib Figure object.
            - npt.NDArray[plt.Axes]: An array of Axes objects representing the subplots.

        See Also
        --------
        CADETProcess.optimization.Population.plot_pareto
        """
        plot = None
        _show = False
        _plot_directory = None

        cNorm = colors.Normalize(vmin=0, vmax=self.n_gen)
        scalarMap_feas = cmx.ScalarMappable(norm=cNorm, cmap=cmap_feas)
        scalarMap_infeas = cmx.ScalarMappable(norm=cNorm, cmap=cmap_infeas)

        if plot_pareto:
            populations = self.pareto_fronts
            population_last = self.pareto_front
        else:
            populations = self.populations
            population_last = self.population_last

        if not plot_evolution:
            populations = [population_last]

        for i, gen in enumerate(populations):
            if gen is population_last:
                _plot_directory = plot_directory
                _show = show
            plot = gen.plot_pareto(
                plot,
                color_feas=scalarMap_feas.to_rgba(i),
                color_infeas=scalarMap_infeas.to_rgba(i),
                show=_show,
                plot_directory=_plot_directory,
            )

        return plot.fig, plot.ax

    @wraps(Population.plot_corner)
    def plot_corner(self, *args: Any, **kwargs: Any) -> None:
        """Create corner plot of population."""
        return self.population_all.plot_corner(*args, **kwargs)

    @wraps(Population.plot_pairwise)
    def plot_pairwise(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> tuple[plt.Figure, np.ndarray[plt.Axes]]:
        """Plot population pairwise."""
        return self.population_all.plot_pairwise(*args, **kwargs)

    def setup_convergence_figure(
        self,
        target: Literal["objectives", "nonlinear_constraints", "meta_scores"],
        plot_individual: bool = False,
    ) -> tuple[list, list]:
        """
        Set up figures and axes for plotting convergence of specified targets.

        Parameters
        ----------
        target : str
            The target type for convergence plotting. Options are "objectives",
            "nonlinear_constraints", or "meta_scores".
        plot_individual : bool, optional
            If True, individual figures are created for each target. Otherwise, a single
            figure with subplots is created. Default is False.

        Returns
        -------
        tuple[list, list]
            A tuple containing lists of matplotlib Figure and Axes objects.
            Returns individual figures and axes if `plot_individual` is True.

        Raises
        ------
        CADETProcessError
            If the specified target is unknown or not supported.
        """
        if target == "objectives":
            n = self.optimization_problem.n_objectives
        elif target == "nonlinear_constraints":
            n = self.optimization_problem.n_nonlinear_constraints
        elif target == "meta_scores":
            n = self.optimization_problem.n_meta_scores
        else:
            raise CADETProcessError("Unknown target.")

        if n == 0:
            return (None, None)

        fig_all, axs_all = plt.subplots(
            ncols=n,
            figsize=(n * 6 + 2, 6),
            squeeze=False,
        )
        axs_all = axs_all.reshape((-1,))

        plt.close(fig_all)

        figs_ind = []
        axs_ind = []
        for i in range(n):
            fig, ax = plt.subplots()
            figs_ind.append(fig)
            axs_ind.append(ax)
            plt.close(fig)

        axs_ind = np.array(axs_ind).reshape(axs_all.shape)

        if plot_individual:
            return figs_ind, axs_ind
        else:
            return fig_all, axs_all

    def plot_convergence(
        self,
        target: Literal["objectives", "nonlinear_constraints", "meta_scores"] = "objectives",
        figs: Optional[Figure] = None,
        axs: Optional[Axes] = None,
        plot_individual: bool = False,
        plot_avg: bool = True,
        autoscale: bool = True,
        show: bool = True,
        plot_directory: bool = None,
    ) -> tuple[list[Figure] | Figure, list[Axes]]:
        """
        Plot the convergence of optimization metrics over evaluations.

        Parameters
        ----------
        target : Literal["objectives", "nonlinear_constraints", "meta_scores"],
            The target metrics to plot. The default is "objectives".
        figs : plt.Figure or list of plt.Figure, optional
            Figure(s) to plot the objectives on.
        axs : plt.Axes or list of plt.Axes, optional
            Axes to plot the objectives on.
            If None, new figures and axes will be created.
        plot_individual : bool, optional
            If True, create individual figure vor each metric.
            The default is False.
        plot_avg : bool, optional
            If True, plot add trajectory of average value per generation.
            The default is True.
        autoscale : bool, optional
            If True, autoscale the y-axis. The default is True.
        show : bool, optional
            If True, show the plot. The default is True.
        plot_directory : str, optional
            A directory to save the plot, by default None.

        Returns
        -------
        tuple
            A tuple containing:
            - plt.Figure: The Matplotlib Figure object.
            - npt.NDArray[plt.Axes]: An array of Axes objects representing the subplots.
        """
        if axs is None:
            figs, axs = self.setup_convergence_figure(target, plot_individual)

        if not isinstance(figs, list):
            figs = [figs]

        layout = plotting.Layout()
        layout.x_label = "$n_{Evaluations}$"

        if target == "objectives":
            funcs = self.optimization_problem.objectives
            values_min = self.f_best_history
            values_avg = self.f_avg_history
        elif target == "nonlinear_constraints":
            funcs = self.optimization_problem.nonlinear_constraints
            values_min = self.g_best_history
            values_avg = self.g_avg_history
        elif target == "meta_scores":
            funcs = self.optimization_problem.meta_scores
            values_min = self.m_best_history
            values_avg = self.m_avg_history
        else:
            raise CADETProcessError("Unknown target.")

        if len(funcs) == 0:
            return

        counter = 0
        for func in funcs:
            start = counter
            stop = counter + func.n_metrics
            v_func_min = values_min[:, start:stop]
            v_func_avg = values_avg[:, start:stop]

            for i_metric in range(func.n_metrics):
                v_line_min = v_func_min[:, i_metric]
                v_line_avg = v_func_avg[:, i_metric]

                ax = axs[counter + i_metric]
                lines = ax.get_lines()

                if len(lines) > 0:
                    lines[0].set_xdata(self.n_evals_history)
                    lines[0].set_ydata(v_line_min)
                    if plot_avg and self.population_last.n_individuals > 1:
                        lines[1].set_ydata(v_line_avg)
                else:
                    if plot_avg and self.population_last.n_individuals > 1:
                        label = "best"
                    else:
                        label = None

                    ax.plot(
                        self.n_evals_history, v_line_min, "--", color="k", label=label
                    )
                    if plot_avg and self.population_last.n_individuals > 1:
                        ax.plot(
                            self.n_evals_history,
                            v_line_avg,
                            "-",
                            color="k",
                            alpha=0.5,
                            label="avg",
                        )

                layout.x_lim = (0, np.max(self.n_evals_history) + 1)

                try:
                    label = func.labels[i_metric]
                except AttributeError:
                    label = f"{func}_{i_metric}"

                if plot_avg and self.population_last.n_individuals > 1:
                    y_min = np.nanmin(v_line_min)
                    y_max = np.nanmax(v_line_avg)
                else:
                    y_min = np.nanmin(v_line_min)
                    y_max = np.nanmax(v_line_min)

                layout.y_label = label
                if autoscale and y_min > 0:
                    if y_max / y_min > 100.0:
                        ax.set_yscale("log")

                try:
                    plotting.set_layout(ax, layout)
                    ax.relim()
                    ax.autoscale_view()
                except ValueError:
                    pass

            counter += func.n_metrics

        for fig in figs:
            fig.tight_layout()
            if not show:
                plt.close(fig)
            else:
                dummy = plt.figure(figsize=fig.get_size_inches())
                new_manager = dummy.canvas.manager
                new_manager.canvas.figure = fig
                fig.set_canvas(new_manager.canvas)
                plt.show()

        if plot_directory is not None:
            plot_directory = Path(plot_directory)
            if plot_individual:
                for i, fig in enumerate(figs):
                    figname = f"convergence_{target}_{i}"
                    fig.savefig(f"{plot_directory / figname}.png")
            else:
                figname = f"convergence_{target}"
                figs[0].savefig(f"{plot_directory / figname}.png")

        if plot_individual:
            return figs, axs
        else:
            return figs[0], axs

    def save_results(self, file_name: str) -> None:
        """
        Save results to H5 file.

        Parameters
        ----------
        file_name : str
            Results file name without file extension.
        """
        if self.results_directory is not None:
            self._update_csv(self.population_last, "results_all", mode="a")
            self._update_csv(self.population_last, "results_last", mode="w")
            self._update_csv(self.pareto_front, "results_pareto", mode="w")
            if self.optimization_problem.n_meta_scores > 0:
                self._update_csv(self.meta_front, "results_meta", mode="w")

            results = H5()
            results.root = Dict(self.to_dict())
            results.filename = self.results_directory / f"{file_name}.h5"
            results.save()

    def load_results(self, file_name: str) -> None:
        """
        Update optimization results from an HDF5 checkpoint file.

        Parameters
        ----------
        file_name : str
            Path to the checkpoint file.
        """
        data = H5()
        data.filename = file_name

        # Check for CADET-Python >= v1.1, which introduced the .load_from_file interface.
        # If it's not present, assume CADET-Python <= 1.0.4 and use the old .load() interface
        # This check can be removed at some point in the future.
        if hasattr(data, "load_from_file"):
            data.load_from_file()
        else:
            data.load()

        self.update_from_dict(data.root)

    def to_dict(self) -> dict:
        """
        Convert Results to a dictionary.

        Returns
        -------
        addict.Dict
            Results as a dictionary with populations stored as list of dictionaries.
        """
        data = Dict()
        data.system_information = self.system_information
        data.optimizer_state = self.optimizer_state
        data.similarity_tol = self._similarity_tol
        data.population_all_id = str(self.population_all.id)
        data.populations = {i: pop.to_dict() for i, pop in enumerate(self.populations)}
        data.pareto_fronts = {
            i: front.to_dict() for i, front in enumerate(self.pareto_fronts)
        }
        if self._meta_fronts is not None:
            data.meta_fronts = {
                i: front.to_dict() for i, front in enumerate(self.meta_fronts)
            }
        if self.time_elapsed is not None:
            data.time_elapsed = self.time_elapsed
            data.cpu_time = self.cpu_time

        return data

    def update_from_dict(self, data: dict) -> None:
        """
        Update internal state from dictionary.

        Parameters
        ----------
        data : dict
            Serialized data.
        """
        self._optimizer_state = data["optimizer_state"]
        self._population_all = Population(id=data["population_all_id"])
        self._similarity_tol = data.get("similarity_tol")

        for pop_dict in data["populations"].values():
            pop = Population.from_dict(pop_dict)
            self.update(pop)

        self._pareto_fronts = [
            ParetoFront.from_dict(d) for d in data["pareto_fronts"].values()
        ]
        if self._meta_fronts is not None:
            self._meta_fronts = [
                ParetoFront.from_dict(d) for d in data["meta_fronts"].values()
            ]

    def setup_csv(self) -> None:
        """Create csv files for optimization results."""
        self._setup_csv("results_all")
        self._setup_csv("results_last")
        self._setup_csv("results_pareto")
        if self.optimization_problem.n_meta_scores > 0:
            self._setup_csv("results_meta")

    def _setup_csv(self, file_name: str) -> None:
        """
        Create csv file for optimization results.

        Parameters
        ----------
        file_name : str
            Results file name without file extension.
        """
        header = [
            "id",
            *self.optimization_problem.variable_names,
            *self.optimization_problem.objective_labels,
        ]

        if self.optimization_problem.n_nonlinear_constraints > 0:
            header += [*self.optimization_problem.nonlinear_constraint_labels]
        if self.optimization_problem.n_meta_scores > 0:
            header += [*self.optimization_problem.meta_score_labels]

        with open(f"{self.results_directory / file_name}.csv", "w") as csvfile:
            writer = csv.writer(csvfile, delimiter=",")
            writer.writerow(header)

    def _update_csv(
        self,
        population: Population,
        file_name: str,
        mode: Literal["w", "b"],
    ) -> None:
        """
        Update csv file with latest population.

        Parameters
        ----------
        population : Population
            latest Population.
        file_name : str
            Results file name without file extension.
        mode : {'a', 'w'}
            a: append to existing file.
            w: Create new csv.

        See Also
        --------
        setup_csv
        """
        if mode == "w":
            self._setup_csv(file_name)
            mode = "a"

        with open(f"{self.results_directory / file_name}.csv", mode) as csvfile:
            writer = csv.writer(csvfile, delimiter=",")

            for ind in population:
                row = [ind.id, *ind.x.tolist(), *ind.f.tolist()]
                if ind.g is not None:
                    row += ind.g.tolist()
                if ind.m is not None:
                    row += ind.m.tolist()
                writer.writerow(row)



"""
=================================
Tools (:mod:`CADETProcess.tools`)
=================================

.. currentmodule:: CADETProcess.tools

A collection of tools extending the functionatlity of **CADET-Process**.

Yamamoto's Method
=================

A module to fit isotherm parameters using Yamamoto's method.

.. autosummary::
   :toctree: generated/

   yamamoto

"""



import hashlib
from typing import Any


def digest_string(obj: Any, length: int = 8) -> str:
    """
    Compute a short digest of a string.

    Parameters
    ----------
    obj : Any
        Input to digest. Will be converted to string.
    length : int, optional
        Length of the returned digest (default is 8).

    Returns
    -------
    digest : str
        Hexadecimal string digest of the input.

    Examples
    --------
    >>> digest_string("hello world")
    'b94d27b9'
    """
    s = str(obj)
    return hashlib.sha256(s.encode()).hexdigest()[:length]



import datetime
import importlib
import multiprocessing
import shutil
import time
import unittest

from CADETProcess import settings
from CADETProcess.optimization import U_NSGA3, SequentialBackend
from CADETProcess.simulator import Cadet

from tests.test_cadet_adapter import detect_cadet
from tests.test_optimization_problem import setup_optimization_problem

parallel_backends_module = importlib.import_module("CADETProcess.optimization")

_parallel_backends = [
    "Joblib",
    "Pathos",
]

parallel_backends = []
for Backend in _parallel_backends:
    try:
        Backend = getattr(parallel_backends_module, Backend)
        parallel_backends.append(Backend)
    except AttributeError:
        print(f"Couldn't import backend {Backend}")

backends = [SequentialBackend] + parallel_backends

found_cadet, install_path = detect_cadet()

n_cores = 2
cpu_count = multiprocessing.cpu_count()


class TestParallelizationBackend(unittest.TestCase):
    """Test initializing parallelization backends and n_cores attribute."""

    def test_n_cores(self):
        with self.assertRaises(ValueError):
            sequential_backend = SequentialBackend(n_cores=n_cores)

        with self.assertRaises(ValueError):
            sequential_backend = SequentialBackend()
            sequential_backend.n_cores = n_cores

        for Backend in parallel_backends:
            backend = Backend(n_cores=n_cores)
            backend.n_cores = n_cores

    def test_max_cores(self):
        for Backend in parallel_backends:
            with self.assertRaises(ValueError):
                backend = Backend(n_cores=cpu_count + 1)

    def test_negative_n_cores(self):
        for Backend in parallel_backends:
            backend = Backend(n_cores=0)
            self.assertEqual(backend._n_cores, cpu_count)

            backend.n_cores = -1
            self.assertEqual(backend._n_cores, cpu_count)

            backend.n_cores = -2
            if cpu_count > 1:
                self.assertEqual(backend._n_cores, cpu_count - 1)
            else:
                self.assertEqual(backend._n_cores, 1)

            with self.assertRaises(ValueError):
                backend.n_cores = -cpu_count - 1


class TestParallelEvaluation(unittest.TestCase):
    """Test evaluating function in parallel.

    Note
    ----
    Since initializing CADET in parallel can be an issue, this is also tested.

    """

    def tearDown(self):
        shutil.rmtree("./tmp", ignore_errors=True)
        shutil.rmtree("./test_parallelization", ignore_errors=True)

    def test_parallelization_backend(self):
        def evaluation_function(sleep_time=0.0):
            print(sleep_time)
            time.sleep(sleep_time)
            return True

        for Backend in backends:
            backend = Backend()
            if not isinstance(backend, SequentialBackend):
                backend.n_cores = n_cores
            results = backend.evaluate(evaluation_function, [0.01] * 4)
            self.assertTrue(all(results))

    @unittest.skipIf(found_cadet is False, "Skip if CADET is not installed.")
    def test_parallel_cadet_initialization(self):
        def evaluation_function(x):
            simulator = Cadet()

            file_name = simulator.get_tempfile_name()
            cwd = simulator.temp_dir
            sim = simulator.create_lwe(cwd / file_name)
            if hasattr(sim, "run_simulation"):
                return_information = sim.run_simulation()
            else:
                return_information = sim.run_load()

            return True

        for Backend in backends:
            backend = Backend()
            if not isinstance(backend, SequentialBackend):
                backend.n_cores = n_cores

            results = backend.evaluate(evaluation_function, [0.01] * 4)

            self.assertTrue(all(results))


class TestOptimizerParallelizationBackend(unittest.TestCase):
    """Test parallel backends in optimizer.

    Note
    ----
    Consider moving this to Optimizer tests.

    """

    def tearDown(self):
        settings.working_directory = None

        shutil.rmtree("./test_parallelization", ignore_errors=True)
        shutil.rmtree("./diskcache_simple", ignore_errors=True)

    def test_parallel_optimization(self):
        def run_optimization(backend=None):
            def dummy_objective_function(x):
                y = sum((x - 0.5) ** 2)
                time.sleep(0.05)
                return y

            settings.working_directory = "./test_parallelization"

            optimization_problem = setup_optimization_problem(
                use_diskcache=True, obj_fun=dummy_objective_function
            )

            optimizer = U_NSGA3()
            optimizer.pop_size = 16
            optimizer.n_max_gen = 1

            if backend is not None:
                optimizer.parallelization_backend = backend
                backend_string = str(backend)
            else:
                backend_string = "Default backend"

            start_time = datetime.datetime.now()
            opt_results = optimizer.optimize(
                optimization_problem,
                save_results=True,
                use_checkpoint=False,
            )
            end_time = datetime.datetime.now()

            print(
                f"======> {backend_string} with "
                f"{optimizer.parallelization_backend.n_cores} cores took "
                f"{end_time - start_time} <=========="
            )

        run_optimization(backend=None)

        for Backend in backends:
            backend = Backend()
            if not isinstance(backend, SequentialBackend):
                backend.n_cores = n_cores
            run_optimization(backend=backend)


if __name__ == "__main__":
    unittest.main()



"""
=========================================
Transform (:mod:`CADETProcess.transform`)
=========================================

.. currentmodule:: CADETProcess.transform

This module provides functionality for transforming data.


.. autosummary::
    :toctree: generated/

    TransformerBase
    NullTransformer
    NormLinearTransformer
    NormLogTransformer
    AutoTransformer

"""  # noqa

from abc import ABC, abstractmethod
from typing import Any, Optional

import matplotlib.pyplot as plt
import numpy as np

from CADETProcess import plotting
from CADETProcess.numerics import round_to_significant_digits


class TransformerBase(ABC):
    """
    Base class for parameter transformation.

    This class provides an interface for transforming an input parameter space to some
    output parameter space.

    Attributes
    ----------
    lb_input : float or np.ndarray
        Lower bounds of the input parameter space.
    ub_input : float or np.ndarray
        Upper bounds of the input parameter space.
    lb : float or np.ndarray
        Lower bounds of the output parameter space.
    ub : float or np.ndarray
        Upper bounds of the output parameter space.
    allow_extended_input : bool
        If True, the input value may exceed the lower/upper bounds.
        Else, an exception is thrown.
    allow_extended_output : bool
        If True, the output value may exceed the lower/upper bounds.
        Else, an exception is thrown.

    Raises
    ------
    ValueError
        If lb_input and ub_input have different shapes.

    Notes
    -----
    - This is an abstract base class and cannot be instantiated directly.
    - The `transform` method must be implemented by a subclass.

    Examples
    --------
    >>> class MyTransformer(TransformerBase):
    ...     def transform(self, x: float) -> float:
    ...         return x ** 2
    ...
    >>> t = MyTransformer(lb_input=0, ub_input=10, lb=-100, ub=100)
    >>> t.transform(3)
    9
    """

    def __init__(
        self,
        lb_input: float | np.ndarray = -np.inf,
        ub_input: float | np.ndarray = np.inf,
        allow_extended_input: Optional[bool] = False,
        allow_extended_output: Optional[bool] = False,
    ) -> None:
        """
        Initialize TransformerBase.

        Parameters
        ----------
        lb_input : float or np.ndarray, optional
            Lower bounds of the input parameter space. Default is -inf.
        ub_input : float or np.ndarray, optional
            Upper bounds of the input parameter space. Default is inf.
        allow_extended_input : bool, optional
            If True, the input value may exceed the lower/upper bounds.
            Else, an exception is thrown. Default is False.
        allow_extended_output : bool, optional
            If True, the output value may exceed the lower/upper bounds.
            Else, an exception is thrown. Default is False.
        """
        self.lb_input = lb_input
        self.ub_input = ub_input
        self.allow_extended_input = allow_extended_input
        self.allow_extended_output = allow_extended_output

    @property
    @abstractmethod
    def is_linear(self) -> bool:
        """Return whether the transformation is linear."""
        pass

    @property
    def lb_input(self) -> float | np.ndarray:
        """Return the lower bounds of the input parameter space."""
        return self._lb_input

    @lb_input.setter
    def lb_input(self, lb_input: float | np.ndarray) -> None:
        self._lb_input = lb_input

    @property
    def ub_input(self) -> float | np.ndarray:
        """Return the upper bounds of the input parameter space."""
        return self._ub_input

    @ub_input.setter
    def ub_input(self, ub_input: float | np.ndarray) -> None:
        self._ub_input = ub_input

    @property
    @abstractmethod
    def lb(self) -> float | np.ndarray:
        """Return the lower bounds of the output parameter space."""
        pass

    @property
    @abstractmethod
    def ub(self) -> float | np.ndarray:
        """Return the upper bounds of the output parameter space."""
        pass

    def transform(self, x: float | np.ndarray) -> float | np.ndarray:
        """
        Transform the input parameter space to the output parameter space.

        Applies the transformation function `_transform` to `x` after performing input
        bounds checking. If the transformed value exceeds the output bounds, an error
        is raised.

        Parameters
        ----------
        x : float or np.ndarray
            Input parameter values.

        Returns
        -------
        float or np.ndarray
            Transformed parameter values.

        Raises
        ------
        ValueError
            If `x` exceeds input or output bounds and `allow_extended_*` is False.
        """
        if not self.allow_extended_input and not np.all(
            (self.lb_input <= x) & (x <= self.ub_input)
        ):
            raise ValueError("Value exceeds input bounds.")

        x = self._transform(x)

        if not self.allow_extended_output and not np.all(
            (self.lb <= x) & (x <= self.ub)
        ):
            raise ValueError("Value exceeds output bounds.")

        return x

    @abstractmethod
    def _transform(self, x: float | np.ndarray) -> float | np.ndarray:
        """
        Apply the transformation from input to output parameter space.

        Must be implemented in the child class.

        Parameters
        ----------
        x : float or np.ndarray
            Input parameter values.

        Returns
        -------
        float or np.ndarray
            Transformed parameter values.
        """
        pass

    def untransform(
        self,
        x: float | np.ndarray,
        significant_digits: Optional[int] = None,
    ) -> float | np.ndarray:
        """
        Transform the output parameter space back to the input parameter space.

        Parameters
        ----------
        x : float or np.ndarray
            Output parameter values.
        significant_digits : int, optional
            float | np.ndarray of significant figures to which variable can be rounded.
            If None, variable is not rounded.

        Returns
        -------
        float or np.ndarray
            Transformed parameter values.
        """
        x_ = round_to_significant_digits(x, digits=significant_digits)

        if (
            not self.allow_extended_output
            and
            not np.all((self.lb <= x_) & (x_ <= self.ub))
        ):
            raise ValueError("Value exceeds output bounds.")

        x_ = self._untransform(x_)
        x_ = round_to_significant_digits(x_, digits=significant_digits)

        if (
            not self.allow_extended_input
            and
            not np.all((self.lb_input <= x_) & (x_ <= self.ub_input))
        ):
            raise ValueError("Value exceeds input bounds.")

        return x_

    @abstractmethod
    def _untransform(self, x: float | np.ndarray) -> float | np.ndarray:
        """
        Apply the inverse transformation from output to input parameter space.

        Must be implemented in the child class.

        Parameters
        ----------
        x : float or np.ndarray
            Output parameter values.

        Returns
        -------
        float or np.ndarray
            Transformed parameter values.
        """
        pass

    @plotting.create_and_save_figure
    def plot(self, ax: plt.Axes, use_log_scale: bool = False) -> None:
        """
        Plot the transformed space against the input space.

        Parameters
        ----------
        ax : plt.Axes
            The axes object to plot on.
        use_log_scale : bool, optional
            If True, use a logarithmic scale for the x-axis.
        """
        allow_extended_input = self.allow_extended_input
        self.allow_extended_input = True

        y = np.linspace(self.lb, self.ub)
        x = self.untransform(y)

        ax.plot(x, y)
        ax.set_xlabel("Input Space")
        ax.set_ylabel("Transformed Space")

        if use_log_scale:
            ax.set_xscale("log")

        self.allow_extended_input = allow_extended_input

    def __str__(self) -> str:
        """Return the class name as a string."""
        return self.__class__.__name__


class NullTransformer(TransformerBase):
    """
    A transformer that performs no transformation.

    This class simply returns the input values as output without modification.

    See Also
    --------
    TransformerBase : The base class for parameter transformation.
    """

    @property
    def is_linear(self) -> bool:
        """Return True, as this is a linear transformation."""
        return True

    @property
    def lb(self) -> float | np.ndarray:
        """Return the lower bound of the output space (same as input lower bound)."""
        return self.lb_input

    @property
    def ub(self) -> float | np.ndarray:
        """Return the upper bound of the output space (same as input upper bound)."""
        return self.ub_input

    def _transform(self, x: float | np.ndarray) -> float | np.ndarray:
        """
        Return the input value(s) as output without modification.

        Parameters
        ----------
        x : float or np.ndarray
            The input value(s) to be transformed.

        Returns
        -------
        float or np.ndarray
            The transformed output value(s) (same as input).
        """
        return x

    def _untransform(self, x: float | np.ndarray) -> float | np.ndarray:
        """
        Return the output value(s) as input without modification.

        Parameters
        ----------
        x : float or np.ndarray
            The output value(s) to be untransformed.

        Returns
        -------
        float or np.ndarray
            The untransformed input value(s) (same as output).
        """
        return x


class NormLinearTransformer(TransformerBase):
    """
    A transformer that normalizes values linearly to the range [0, 1].

    This transformation scales the input value between the given lower
    and upper bounds into a normalized range of [0,1].

    See Also
    --------
    TransformerBase : The base class for parameter transformation.
    """

    @property
    def is_linear(self) -> bool:
        """Return True, as this is a linear transformation."""
        return True

    @property
    def lb(self) -> float:
        """Return the lower bound of the output space (0)."""
        return 0.0

    @property
    def ub(self) -> float:
        """Return the upper bound of the output space (1)."""
        return 1.0

    def _transform(self, x: float | np.ndarray) -> float | np.ndarray:
        """
        Normalize input values to the range [0,1].

        Parameters
        ----------
        x : float or np.ndarray
            The input value(s) to be transformed.

        Returns
        -------
        float or np.ndarray
            The transformed output value(s) in the range [0,1].
        """
        return (x - self.lb_input) / (self.ub_input - self.lb_input)

    def _untransform(self, x: float | np.ndarray) -> float | np.ndarray:
        """
        Denormalize output values back to the original range.

        Parameters
        ----------
        x : float or np.ndarray
            The output value(s) in the normalized range [0,1].

        Returns
        -------
        float or np.ndarray
            The untransformed input value(s) in the original range.
        """
        return (self.ub_input - self.lb_input) * x + self.lb_input


class NormLogTransformer(TransformerBase):
    """
    A transformer that normalizes values logarithmically to the range [0, 1].

    This transformation scales input values logarithmically between the given lower
    and upper bounds into a normalized range of [0,1].

    See Also
    --------
    TransformerBase : The base class for parameter transformation.
    """

    @property
    def is_linear(self) -> bool:
        """Return False, as this is a non-linear transformation."""
        return False

    @property
    def lb(self) -> float:
        """Return the lower bound of the output space (0)."""
        return 0.0

    @property
    def ub(self) -> float:
        """Return the upper bound of the output space (1)."""
        return 1.0

    def _transform(self, x: float | np.ndarray) -> float | np.ndarray:
        """
        Normalize input values to the range [0,1] using a logarithmic transformation.

        Parameters
        ----------
        x : float or np.ndarray
            The input value(s) to be transformed.

        Returns
        -------
        float or np.ndarray
            The transformed output value(s) in the range [0,1].

        Raises
        ------
        ValueError
            If `lb_input` is non-positive, the transformation shifts all values accordingly.
        """
        if self.lb_input <= 0:
            x_ = x + (abs(self.lb_input) + 1)
            ub = 1 + (self.ub_input - self.lb_input)
            return np.log(x_) / np.log(ub)
        else:
            return np.log(x / self.lb_input) / np.log(self.ub_input / self.lb_input)

    def _untransform(self, x: float | np.ndarray) -> float | np.ndarray:
        """
        Denormalize output values back to the original range using logarithmic inverse.

        Parameters
        ----------
        x : float or np.ndarray
            The output value(s) in the normalized range [0,1].

        Returns
        -------
        float or np.ndarray
            The untransformed input value(s) in the original range.
        """
        if self.lb_input <= 0:
            return (
                np.exp(x * np.log(self.ub_input - self.lb_input + 1))
                + self.lb_input
                - 1
            )
        else:
            return self.lb_input * np.exp(x * np.log(self.ub_input / self.lb_input))


class AutoTransformer(TransformerBase):
    """
    A transformer that automatically selects between linear and logarithmic transformations.

    Transforms the input value to the range [0, 1] using either
    the :class:`NormLinearTransformer` or the :class:`NormLogTransformer`
    based on the input parameter space.

    Attributes
    ----------
    linear : NormLinearTransformer
        Instance of the linear normalization transform.
    log : NormLogTransformer
        Instance of the logarithmic normalization transform.
    threshold : int
        The maximum threshold to switch from linear to logarithmic transformation.

    See Also
    --------
    TransformerBase
    NormLinearTransformer
    NormLogTransformer
    """

    def __init__(self, *args: Any, threshold: int = 100, **kwargs: Any) -> None:
        """
        Initialize an AutoTransformer object.

        Parameters
        ----------
        *args : tuple
            Arguments for the :class:`TransformerBase` class.
        threshold : int, optional
            The maximum threshold to switch from linear to logarithmic
            transformation. The default is 100.
        **kwargs : dict
            Keyword arguments for the :class:`TransformerBase` class.
        """
        self.linear = NormLinearTransformer()
        self.log = NormLogTransformer()

        super().__init__(*args, **kwargs)
        self.threshold = threshold

        self.linear.allow_extended_input = self.allow_extended_input
        self.linear.allow_extended_output = self.allow_extended_output
        self.log.allow_extended_input = self.allow_extended_input
        self.log.allow_extended_output = self.allow_extended_output

    @property
    def is_linear(self) -> bool:
        """Return True if linear transformation is used, otherwise False."""
        return self.use_linear

    @property
    def use_linear(self) -> bool:
        """Determine whether linear transformation should be used."""
        if self.lb_input <= 0:
            return np.log10(self.ub_input - self.lb_input) < np.log10(self.threshold)
        return (self.ub_input / self.lb_input) < self.threshold

    @property
    def use_log(self) -> bool:
        """Return True if logarithmic transformation is used, otherwise False."""
        return not self.use_linear

    @property
    def lb(self) -> float:
        """Return the lower bound of the output parameter space (0)."""
        return 0.0

    @property
    def ub(self) -> float:
        """Return the upper bound of the output parameter space (1)."""
        return 1.0

    @property
    def lb_input(self) -> float | np.ndarray:
        """Return the lower bounds of the input parameter space."""
        return self._lb_input

    @lb_input.setter
    def lb_input(self, lb_input: float | np.ndarray) -> None:
        """Set the lower bounds of the input parameter space."""
        self.linear.lb_input = lb_input
        self.log.lb_input = lb_input
        self._lb_input = lb_input

    @property
    def ub_input(self) -> float | np.ndarray:
        """Return the upper bounds of the input parameter space."""
        return self._ub_input

    @ub_input.setter
    def ub_input(self, ub_input: float | np.ndarray) -> None:
        """Set the upper bounds of the input parameter space."""
        self.linear.ub_input = ub_input
        self.log.ub_input = ub_input
        self._ub_input = ub_input

    def _transform(self, x: float | np.ndarray) -> float | np.ndarray:
        """
        Transform the input value to an output value in the range [0, 1].

        Parameters
        ----------
        x : float or np.ndarray
            The input value(s) to be transformed.

        Returns
        -------
        float or np.ndarray
            The transformed output value(s).
        """
        return self.log._transform(x) if self.use_log else self.linear._transform(x)

    def _untransform(self, x: float | np.ndarray) -> float | np.ndarray:
        """
        Untransform the output value back to the input parameter space.

        Parameters
        ----------
        x : float or np.ndarray
            The output value(s) in the transformed range.

        Returns
        -------
        float or np.ndarray
            The untransformed input value(s).
        """
        return self.log._untransform(x) if self.use_log else self.linear._untransform(x)



import unittest

import CADETProcess
import numpy as np


class Test_Fractions(unittest.TestCase):
    def create_fractions(self):
        m_0 = np.array([0, 0])
        frac0 = CADETProcess.fractionation.Fraction(m_0, 1)
        m_1 = np.array([3, 0])
        frac1 = CADETProcess.fractionation.Fraction(m_1, 2)
        m_2 = np.array([1, 2])
        frac2 = CADETProcess.fractionation.Fraction(m_2, 3)
        m_3 = np.array([0, 2])
        frac3 = CADETProcess.fractionation.Fraction(m_3, 3)
        m_4 = np.array([0, 0])
        frac4 = CADETProcess.fractionation.Fraction(m_4, 0)

        return frac0, frac1, frac2, frac3, frac4

    def create_pools(self):
        fractions = self.create_fractions()

        pool_waste = CADETProcess.fractionation.FractionPool(n_comp=2)
        pool_waste.add_fraction(fractions[0])
        pool_waste.add_fraction(fractions[1])
        pool_waste.add_fraction(fractions[2])

        pool_1 = CADETProcess.fractionation.FractionPool(n_comp=2)
        pool_1.add_fraction(fractions[3])

        pool_2 = CADETProcess.fractionation.FractionPool(n_comp=2)
        pool_2.add_fraction(fractions[4])

        return pool_waste, pool_1, pool_2

    def test_fraction_mass(self):
        fractions = self.create_fractions()

        self.assertEqual(fractions[0].fraction_mass, 0)
        self.assertEqual(fractions[1].fraction_mass, 3)
        self.assertEqual(fractions[2].fraction_mass, 3)
        self.assertEqual(fractions[3].fraction_mass, 2)
        self.assertEqual(fractions[4].fraction_mass, 0)

    def test_fraction_concentration(self):
        fractions = self.create_fractions()

        np.testing.assert_equal(fractions[0].concentration, np.array([0.0, 0.0]))
        np.testing.assert_equal(fractions[1].concentration, np.array([1.5, 0]))
        np.testing.assert_equal(fractions[2].concentration, np.array([1 / 3, 2 / 3]))
        np.testing.assert_equal(fractions[3].concentration, np.array([0, 2 / 3]))
        np.testing.assert_equal(fractions[4].concentration, np.array([0, 0]))

    def test_fraction_purity(self):
        fractions = self.create_fractions()

        np.testing.assert_equal(fractions[0].purity, np.array([0.0, 0.0]))
        np.testing.assert_equal(fractions[1].purity, np.array([1, 0]))
        np.testing.assert_equal(fractions[2].purity, np.array([1 / 3, 2 / 3]))
        np.testing.assert_equal(fractions[3].purity, np.array([0, 1]))
        np.testing.assert_equal(fractions[4].purity, np.array([0, 0]))

    def test_n_comp(self):
        pools = self.create_pools()

        self.assertEqual(pools[0].n_comp, 2)
        self.assertEqual(pools[0].fractions[0].n_comp, 2)

        m_wrong_n_comp = np.array([0, 0, 0])
        frac_wrong_n_comp = CADETProcess.fractionation.Fraction(m_wrong_n_comp, 1)

        with self.assertRaises(CADETProcess.CADETProcessError):
            pools[0].add_fraction(frac_wrong_n_comp)

    def test_pool_mass(self):
        pools = self.create_pools()

        np.testing.assert_equal(pools[0].mass, np.array([4.0, 2.0]))
        np.testing.assert_equal(pools[1].mass, np.array([0, 2]))
        np.testing.assert_equal(pools[2].mass, np.array([0, 0]))

    def test_pool_pool_mass(self):
        pools = self.create_pools()

        self.assertEqual(pools[0].pool_mass, 6)
        self.assertEqual(pools[1].pool_mass, 2)
        self.assertEqual(pools[2].pool_mass, 0)

    def test_pool_volume(self):
        pools = self.create_pools()

        self.assertEqual(pools[0].volume, 6)
        self.assertEqual(pools[1].volume, 3)
        self.assertEqual(pools[2].volume, 0)

    def test_pool_concentration(self):
        pools = self.create_pools()

        np.testing.assert_equal(pools[0].concentration, np.array([2 / 3, 1 / 3]))
        np.testing.assert_equal(pools[1].concentration, np.array([0, 2 / 3]))
        np.testing.assert_equal(pools[2].concentration, np.array([0, 0]))

    def test_pool_purity(self):
        pools = self.create_pools()

        np.testing.assert_equal(pools[0].purity, np.array([2 / 3, 1 / 3]))
        np.testing.assert_equal(pools[1].purity, np.array([0, 1]))
        np.testing.assert_equal(pools[2].purity, np.array([0, 0]))


if __name__ == "__main__":
    unittest.main()



"""
======================================================
Cyclic Stationarity (:mod:`CADETProcess.stationarity`)
======================================================

.. currentmodule:: CADETProcess.stationarity

Module to evaluate cyclic stationarity of succeeding cycles.

.. autosummary::
    :toctree: generated/

    RelativeArea
    NRMSE
    StationarityEvaluator

"""  # noqa

from typing import Any, Optional

import numpy as np
from addict import Dict

from CADETProcess import SimulationResults, log
from CADETProcess.comparison import Comparator
from CADETProcess.dataStructure import Structure, UnsignedFloat
from CADETProcess.processModel import Inlet

__all__ = ["RelativeArea", "NRMSE", "StationarityEvaluator"]


class CriterionBase(Structure):
    threshold = UnsignedFloat(default=1e-3)

    def __str__(self) -> str:
        return self.__class__.__name__


class RelativeArea(CriterionBase):
    """Class to evaluate difference in relative area as stationarity critereon."""

    pass


class NRMSE(CriterionBase):
    """Class to evaluate NRMSE as stationarity critereon."""

    pass


class StationarityEvaluator(Comparator):
    """Class for checking two succeding chromatograms for stationarity."""

    valid_criteria = ["RelativeArea", "NRMSE"]

    def __init__(
        self,
        criteria: Optional[list[CriterionBase]] = None,
        log_level: str = "WARNING",
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        Initialize the stationarity evaluator.

        Parameters
        ----------
        criteria : List[CriterionBase], optional
            List of criteria for stationarity evaluation, by default None
        log_level : str, optional
            The logging level, by default 'WARNING'
        args : list
            Additional arguments.
        kwargs : dict
            Additional keyword arguments.
        """
        # TODO: Check why cirteria are not stored.
        super().__init__(*args, **kwargs)

        self.logger = log.get_logger("StationarityEvaluator", level=log_level)

        self._criteria = []

    @property
    def criteria(self) -> list:
        """list: List of criteria."""
        return self._criteria

    def add_criterion(self, criterion: CriterionBase) -> None:
        """
        Add a criterion to the list of criteria.

        Parameters
        ----------
        criterion : CriterionBase
            Criterion to add to the list of criteria.
        """
        if not isinstance(criterion, CriterionBase):
            raise TypeError("Expected CriterionBase.")

        self._criteria.append(criterion)

    def assert_stationarity(self, simulation_results: SimulationResults) -> bool:
        """
        Check stationarity of two succeeding cycles.

        Parameters
        ----------
        simulation_results : SimulationResults
            Results of current cycle.

        Returns
        -------
        bool
            True if stationarity is reached. False otherwise.

        Raises
        ------
        TypeError
            If simulation_results is not a SimulationResults object.
        """
        self._metrics = []
        criteria = Dict()
        if not isinstance(simulation_results, SimulationResults):
            raise TypeError("Expcected SimulationResults")

        stationarity = True
        for unit, solution in simulation_results.solution_cycles.items():
            if isinstance(simulation_results.process.flow_sheet[unit], Inlet):
                continue
            solution_previous = solution.outlet[-2]
            solution_this = solution.outlet[-1]
            self.add_reference(solution_previous, update=True, smooth=False)

            for c in self.criteria:
                metric = self.add_difference_metric(
                    str(c), unit, f"{unit}.outlet", smooth=False
                )
                criteria[unit][str(c)]["threshold"] = c.threshold
                diff = metric.evaluate(solution_this)
                criteria[unit][str(c)]["metric"] = diff
                if not np.all(diff <= c.threshold):
                    s = False
                    stationarity = s
                else:
                    s = True
                criteria[unit][str(c)]["stationarity"] = s

        self.logger.debug(f"Stationrity criteria: {criteria}")

        return stationarity



from typing import Callable, Optional

import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt
from scipy.optimize import curve_fit

from CADETProcess import CADETProcessError
from CADETProcess.processModel import (
    ChromatographicColumnBase,
    StericMassAction,
    TubularReactorBase,
)

__all__ = [
    "GradientExperiment",
    "plot_experiments",
    "YamamotoResults",
    "fit_parameters",
]


class GradientExperiment:
    """Gradient Experiment Class."""

    def __init__(
        self,
        time: npt.ArrayLike,
        c_salt: npt.ArrayLike,
        c_protein: npt.ArrayLike,
        gradient_volume: float,
        c_salt_start: Optional[float] = None,
        c_salt_end: Optional[float] = None,
    ) -> None:
        """
        Initialize a GradientExperiment instance.

        Parameters
        ----------
        time : ArrayLike
            Time points in seconds.
        c_salt : ArrayLike
            Salt concentration in mM.
        c_protein : ArrayLike
            Protein concentration(s) in mM.
        gradient_volume : float
            Gradient volume in m³.
        c_salt_start : Optional[float], default=None
            Starting salt concentration in mM.
        c_salt_end : Optional[float], default=None
            Ending salt concentration in mM.
        """
        self.time = np.array(time)
        self.c_salt = np.array(c_salt)

        c_protein = np.array(c_protein)
        if len(c_protein.shape) == 1:
            c_protein = np.expand_dims(c_protein, axis=1)
        self.c_protein = c_protein
        self.gradient_volume = gradient_volume

        if c_salt_start is None:
            c_salt_start = np.min(c_salt)
        self.c_salt_start = c_salt_start

        if c_salt_end is None:
            c_salt_end = np.max(c_salt)
        self.c_salt_end = c_salt_end

    @property
    def n_proteins(self) -> int:
        """int: Number of proteins."""
        return self.c_protein.shape[1]

    @property
    def c_salt_at_max(self) -> float:
        """float: Salt concentration at protein peak maximum in mM."""
        max_c_protein = np.argmax(self.c_protein, axis=0)

        return self.c_salt[max_c_protein]

    @property
    def t_at_max(self) -> float:
        """int: Number of proteins."""
        max_c_protein = np.argmax(self.c_protein, axis=0)

        return self.time[max_c_protein]

    def calculate_normalized_gradient_slope(
        self,
        column_volume: float,
        total_porosity: float,
    ) -> float:
        """
        Calculate normalized concentration gradient slope.

        Parameters
        ----------
        column_volume : float
            Void volume of column in m³.
        total_porosity : float
            Total porosity of column.

        Returns
        -------
        normalized_slope : float
            Gradient slope in mM.
        """
        slope = (self.c_salt_end - self.c_salt_start) / self.gradient_volume
        vol_factor = column_volume - total_porosity * column_volume
        return slope * vol_factor

    def plot(
        self,
        fig: Optional[plt.Figure] = None,
        ax: Optional[plt.Axes] = None,
        sec_ax: Optional[plt.Axes] = None,
    ) -> tuple[plt.Figure, plt.Axes, plt.Axes]:
        """
        Plot the gradient experiment data.

        Parameters
        ----------
        fig : Optional[plt.Figure], default=None
            Existing matplotlib Figure object.
        ax : Optional[plt.Axes], default=None
            Existing matplotlib Axes object for primary plot.
        sec_ax : Optional[plt.Axes], default=None
            Existing secondary Axes for salt concentration.

        Returns
        -------
        fig : plt.Figure
            The figure object containing the plot.
        ax : plt.Axes
            The axes object for the protein concentration.
        sec_ax : plt.Axes
            The secondary axes object for the salt concentration.
        """
        if ax is None:
            fig, ax = plt.subplots()

            ax.set_xlabel("$Time / s$")
            ax.set_ylabel("$c_{Protein} / mM$")

            sec_ax = ax.twinx()
            sec_ax.set_ylabel("$c_{Salt} / mM$")

        ax.plot(self.time, self.c_protein, label="Protein")

        c_p_max = np.max(self.c_protein, axis=0)
        ax.vlines(self.t_at_max, ymin=0, ymax=c_p_max, color="k", linestyle="--")
        sec_ax.plot(self.time, self.c_salt, "k", label="Salt")
        sec_ax.plot(self.t_at_max, self.c_salt_at_max, "ro")

        fig.tight_layout()

        return fig, ax, sec_ax


def plot_experiments(experiments: list[GradientExperiment]) -> None:
    """
    Plot multiple gradient experiments in a single figure.

    Parameters
    ----------
    experiments : list of GradientExperiment
        List of gradient experiment instances to plot.
    """
    fig = ax = sec_ax = None

    for exp in experiments:
        fig, ax, sec_ax = exp.plot(fig, ax, sec_ax)


def yamamoto_equation(
    log_c_salt_at_max_M: float,
    lambda_: float,
    nu: float,
    k_eq: float,
) -> np.ndarray:
    r"""
    Calculate the theoretical normalized gradient slope using Yamamoto's method.

    Yamamoto's method is used in ion-exchange chromatography to model the relationship
    between the normalized gradient slope (GH) and the peak salt concentration (I_R)
    during a linear salt gradient elution. This method is based on the steric mass
    action (SMA) model and allows for the determination of key chromatographic
    parameters such as the characteristic charge (ν) and the equilibrium constant
    (k_eq).

    The underlying equation is:

    .. math::
        \log(GH) = (ν + 1) \log(I_R) - \log(k_{eq} \cdot \lambda^ν \cdot (ν + 1))

    where:
        - GH: Normalized gradient slope (logarithmic scale)
        - I_R: Salt concentration at the protein peak maximum (in M)
        - ν (nu): Characteristic charge of the protein
        - k_eq: Equilibrium constant (binding affinity)
        - λ (lambda_): Resin capacity in mM

    Parameters
    ----------
    log_c_salt_at_max_M : np.ndarray
        Log10 of salt concentrations at protein peak maximum in M.
    lambda_ : float
        Resin capacity in mM.
    nu : float
        Characteristic charge of the molecule.
    k_eq : float
        Equilibrium constant of the binding model.

    Returns
    -------
    np.ndarray
        Calculated normalized gradient slope (GH) values in logarithmic scale.
    """
    lambda_M = lambda_ / 1000

    return np.multiply((nu + 1), log_c_salt_at_max_M) - np.log10(
        k_eq * lambda_M**nu * (nu + 1)
    )


class YamamotoResults:
    """Parameter values determined using Yamamoto's method."""

    def __init__(
        self,
        column: ChromatographicColumnBase,
        experiments: list[GradientExperiment],
        log_gradient_slope: npt.ArrayLike,
        log_c_salt_at_max_M: npt.ArrayLike,
    ) -> None:
        """
        Initialize YamamotoResults with column, experiments, and log-transformed data.

        Parameters
        ----------
        column : Column
            Column object containing the binding model.
        experiments : list of GradientExperiment
            List of gradient experiment instances.
        log_gradient_slope : np.ndarray
            Normalized gradient slopes in logarithmic scale.
        log_c_salt_at_max_M : np.ndarray
            Log10 of salt concentrations at protein peak maximum in M.
        """
        self.column = column
        self.experiments = experiments
        self.log_gradient_slope = log_gradient_slope
        self.log_c_salt_at_max_M = log_c_salt_at_max_M

    @property
    def characteristic_charge(self) -> np.ndarray:
        """np.ndarray: Characteristic charges of the binding model."""
        return np.array(self.column.binding_model.characteristic_charge[1:])

    @property
    def k_eq(self) -> np.ndarray:
        """np.ndarray: Equilibrium constants of the binding model."""
        return np.array(self.column.binding_model.adsorption_rate[1:])

    def plot(
        self, fig: Optional[plt.Figure] = None, ax: Optional[plt.Axes] = None
    ) -> tuple[plt.Figure, plt.Axes]:
        """
        Plot the normalized gradient slope against the peak salt concentration.

        Parameters
        ----------
        fig : Optional[plt.Figure], default=None
            Existing matplotlib Figure object.
        ax : Optional[plt.Axes], default=None
            Existing matplotlib Axes object.

        Returns
        -------
        fig : plt.Figure
            The figure object containing the plot.
        ax : plt.Axes
            The axes object for the plot.
        """
        if ax is None:
            fig, ax = plt.subplots()
            ax.set_ylabel("Normalized Gradient Slope $GH$ / $M$")
            ax.set_xlabel("Peak Salt Concentration $I_R$ / $M$")

        n_proteins = self.experiments[0].n_proteins

        for i_p in range(n_proteins):
            k_eq = self.k_eq[i_p]
            nu = self.characteristic_charge[i_p]

            x = [
                min(self.log_c_salt_at_max_M[:, i_p]) * 1.05,
                max(self.log_c_salt_at_max_M[:, i_p]) * 0.95,
            ]

            y = yamamoto_equation(
                log_c_salt_at_max_M=x,
                lambda_=self.column.binding_model.capacity,
                nu=nu,
                k_eq=k_eq,
            )
            ax.plot(x, y, "k")
            ax.plot(self.log_c_salt_at_max_M[:, i_p], self.log_gradient_slope, "ro")

        fig.tight_layout()
        return fig, ax


def fit_parameters(
    experiments: list,
    column: ChromatographicColumnBase,
) -> YamamotoResults:
    """
    Fit parameters using Yamamoto's method.

    Parameters
    ----------
    experiments : list
        List of GradientExperiments.
    column : Column
        Column object with binding model

    Returns
    -------
    yamamoto_results : YamamotoResults
        Parameter values.
    """
    if not isinstance(column, TubularReactorBase):
        raise TypeError("Expected Column Model.")

    if not isinstance(column.binding_model, StericMassAction):
        raise TypeError("Yamamoto's method only works for Steric Mass Action model.")

    n_proteins = column.n_comp - 1
    for exp in experiments:
        if exp.n_proteins != n_proteins:
            raise CADETProcessError("Experiments must have same number of proteins.")

    normalized_gradient_slope = [
        exp.calculate_normalized_gradient_slope(column.volume, column.total_porosity)
        for exp in experiments
    ]
    log_gradient_slope = np.log10(np.divide(normalized_gradient_slope, 1000))

    nu = np.zeros((experiments[0].n_proteins,))
    k_eq = np.zeros((experiments[0].n_proteins,))
    log_c_salt_at_max_M = np.zeros(
        (
            len(experiments),
            experiments[0].n_proteins,
        )
    )

    for i_p in range(experiments[0].n_proteins):
        c_salt_at_max = [exp.c_salt_at_max[i_p] for exp in experiments]
        log_c_salt_at_max_M[:, i_p] = np.log10(np.array(c_salt_at_max) / 1000)

        nu[i_p], k_eq[i_p] = _fit_yamamoto(
            log_c_salt_at_max_M[:, i_p],
            log_gradient_slope,
            column.binding_model.capacity,
        )

    column.binding_model.characteristic_charge = [0, *nu.tolist()]
    column.binding_model.adsorption_rate = [0, *k_eq.tolist()]

    yamamoto_results = YamamotoResults(
        column, experiments, log_gradient_slope, log_c_salt_at_max_M
    )

    return yamamoto_results


def _fit_yamamoto(
    log_c_salt_at_max_M: np.ndarray, log_gradient_slope: np.ndarray, lambda_: float
) -> tuple[float, float]:
    """
    Fit the Yamamoto model to experimental data using non-linear curve fitting.

    Parameters
    ----------
    log_c_salt_at_max_M : np.ndarray
        Log10 of salt concentrations at protein peak maximum in mM.
    log_gradient_slope : np.ndarray
        Normalized gradient slopes in logarithmic scale.
    lambda_ : float
        Resin capacity in mM.

    Returns
    -------
    nu : float
        Fitted characteristic charge.
    k_eq : float
        Fitted equilibrium constant.
    """
    bounds = ((0, 1e-10), (1000, 1000))

    def yamamoto_wrapper(c_s: float, nu: float, k_eq: float) -> Callable:
        return yamamoto_equation(c_s, lambda_, nu, k_eq)

    results, pcov = curve_fit(
        yamamoto_wrapper,
        log_c_salt_at_max_M,
        log_gradient_slope,
        bounds=bounds,
        p0=(1, 1),
    )

    return results



from __future__ import annotations

from abc import ABC, ABCMeta
from collections import OrderedDict
from functools import wraps
from inspect import Parameter, Signature
from typing import Any, Callable, Iterable, Iterator, Optional, Type
from warnings import warn

from addict import Dict


# %% Descriptors
class Descriptor(ABC):
    """
    Base class for descriptors.

    Descriptors are used to efficiently implement class attributes that
    require checking type, value, size etc.
    For using Descriptors, a class must inherit from StructMeta.

     - ``self`` is the Descriptor managing the attribute of the ``instance``.
     - ``instance`` is the object which holds the actual ``value``.
     - ``value`` is the value of the ``instance`` attribute.

    See Also
    --------
    StructMeta
    Parameters
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        pass

    def __get__(self, instance: Any, owner: Optional[type[Any]]) -> Any:
        """
        Get the attribute value from the instance's dictionary.

        Parameters
        ----------
        instance : Any
            The instance accessing the attribute.
        owner : Optional[type[Any]]
            The owner class of the descriptor.

        Returns
        -------
        Any
            The attribute value from the instance's dictionary.
        """
        return instance.__dict__[self.name]

    def __set__(self, instance: Any, value: Any) -> None:
        """
        Set the attribute value in the instance's dictionary.

        Parameters
        ----------
        instance : Any
            The instance to set the attribute on.
        value : Any
            The value to set.
        """
        if value is None:
            try:
                del instance.__dict__[self.name]
            except KeyError:
                pass
            return
        instance.__dict__[self.name] = value

    def __delete__(self, instance: Any) -> None:
        """
        Delete the attribute from the instance's dictionary.

        Parameters
        ----------
        instance : Any
            The instance to delete the attribute from.
        """
        del instance.__dict__[self.name]


class ProxyList:
    """A proxy list that dynamically updates attributes of container elements."""

    def __init__(self, aggregator: Aggregator, instance: Any) -> None:
        """Initialize Proxy List."""
        self.aggregator = aggregator
        self.instance = instance

    def _get_values_from_aggregator(self) -> Any:
        """Fetch the latest values from the aggregator."""
        return self.aggregator._get_values_from_container(self.instance, check=True)

    def __getitem__(self, index: int) -> Any:
        """Retrieve an item from the aggregated parameter list (live view)."""
        return self._get_values_from_aggregator()[index]

    def __setitem__(self, index: int, value: Any) -> None:
        """
        Modify an individual element in the aggregated parameter list.

        Ensures changes propagate to the underlying objects.
        """
        current_value = self._get_values_from_aggregator()
        current_value[index] = value
        self.aggregator.__set__(self.instance, current_value)

    def __iter__(self) -> Iterator:
        """Iterate over aggregated values."""
        return iter(self._get_values_from_aggregator())

    def __len__(self) -> int:
        """Return the length of the container."""
        return len(self._get_values_from_aggregator())

    def __repr__(self) -> str:
        """str: String representation for debugging."""
        return f"ProxyList({self._get_values_from_aggregator().__repr__()})"

    def __eq__(self, other: ProxyList) -> bool:
        """Equality comparison."""
        return list(self._get_values_from_aggregator()) == other


class Aggregator:
    """Descriptor aggregating parameters from iterable container of other objects."""

    def __init__(
        self,
        parameter_name: str,
        container: str,
        *args: dict,
        **kwargs: dict,
    ) -> None:
        """
        Initialize the Aggregator.

        Parameters
        ----------
        parameter_name : str
            Name of the parameter to be aggregated.
        container : str
            Name of the iterable attribute in the instance that contains the other
            objects from which parameters will be aggregated.
        *args : tuple, optional
            Additional positional arguments.
        **kwargs : dict, optional
            Additional keyword arguments.
        """
        self.parameter_name = parameter_name
        self.container = container

    def _container_obj(self, instance: Any) -> Iterable:
        """
        Retrieve the iterable container of the instance.

        Parameters
        ----------
        instance : Any
            Instance to retrieve the container from.

        Returns
        -------
        obj : iterable
            Iterable container of the instance.

        Raises
        ------
        TypeError
            If the container is not iterable.
        """
        container = getattr(instance, self.container)

        if not hasattr(container, "__iter__"):
            raise TypeError(f"{self.container} attribute is not iterable")

        return container

    def _n_instances(self, instance: Any) -> int:
        return len(self._container_obj(instance))

    def _get_values_from_container(self, instance: Any, check: bool = False) -> Any:
        container = self._container_obj(instance)

        value = [getattr(el, self.parameter_name) for el in container]

        if check:
            value = self._prepare(instance, value, recursive=True)
            self._check(instance, value, recursive=True)

        return value

    def __get__(self, instance: Any, cls: Type) -> ProxyList:
        """
        Retrieve the descriptor value for the given instance.

        Parameters
        ----------
        instance : Any
            Instance to retrieve the descriptor value for.
        cls : type[Any], optional
            Class to which the descriptor belongs. By default None.

        Returns
        -------
        np.array
            Descriptor values aggregated in a numpy array.
        """
        if instance is None:
            return self

        return ProxyList(self, instance)

    def __set__(self, instance: Any, value: Iterable) -> None:
        """
        Set the descriptor value for the given instance.

        Parameters
        ----------
        instance : Any
            Instance to set the descriptor value for.
        value : Iterable
            Value to set. Note, this assumes that each element of the value maps to
            each element of the container.
        """
        if value is not None:
            value = self._prepare(instance, value, recursive=True)
            self._check(instance, value, recursive=True)

        container = self._container_obj(instance)

        for el, el_value in zip(container, value):
            setattr(el, self.parameter_name, el_value)

    def _prepare(self, instance: Any, value: Any, recursive: bool = False) -> Any:
        """
        Prepare value for setting if necessary.

        Override this method if type-casting or other operations are necessary.

        Parameters
        ----------
        instance : Any
            Instance to retrieve the descriptor value for.
        value : Any
            Value to cast.
        recursive : bool, optional
            If True, perform the check recursively. Defaults to False. Only works in
            case of overriding.

        Returns
        -------
        Any
            Prepared value.
        """
        return value

    def _check(self, instance: Any, value: Iterable, recursive: bool = False) -> None:
        """
        Check the given value.

        Override this method for specific checks.

        Parameters
        ----------
        instance : Any
            Instance to retrieve the descriptor value for.
        value : Iterable
            Value to set. Note, this assumes that each element of the value maps to
            each element of the container.
        recursive : bool, optional
            If True, perform the check recursively. Defaults to False.
        """
        container = self._container_obj(instance)

        if len(value) != len(container):
            raise ValueError(
                f"Unexpected length. Expected {len(container)} entries, got {len(value)}."
            )

        return


def make_signature(names: list[str]) -> Signature:
    """
    Create a signature object from a list of parameter names.

    Parameters
    ----------
    names : list[str]
        List of parameter names.

    Returns
    -------
    Signature
        A Signature object for the given parameter names.
    """
    return Signature(Parameter(name, Parameter.POSITIONAL_OR_KEYWORD) for name in names)


class StructMeta(type):
    """
    Metaclass for creating classes that use Descriptors.

    This metaclass enables classes to have ordered descriptors and provides
    additional functionality related to descriptor management. The underlying
    structure uses an OrderedDict to maintain the order of class attributes.

    The metaclass mainly interacts with the `Descriptor` class, and classes
    that use this metaclass can benefit from this specialized handling of descriptors.

    Attributes
    ----------
    _descriptors : list
        List of descriptors associated with a class.
    _parameters : list
        List of parameters aggregated from the class and its bases.
    _sized_parameters : list
        List of parameters that have a `size` attribute.
    _polynomial_parameters : list
        List of parameters with `fill_values` attribute.
    _aggregated_parameters : list
        List of parameters that aggregate other instances parameters.
    _required_parameters : list
        List of parameters that have a default value of None.

    See Also
    --------
    Structure : Base class that typically uses this metaclass.
    Descriptor : Class that represents the descriptors this metaclass operates on.
    Parameters : Base class for model parameters with e.g. type or bound constraints.

    Methods
    -------
    __prepare__(name, bases) -> OrderedDict
        Prepares the namespace for the class body to be executed.
    """

    @classmethod
    def __prepare__(cls, name: str, bases: tuple) -> OrderedDict:
        """
        Prepare the class namespace using an OrderedDict.

        Parameters
        ----------
        name : str
            Name of the class being defined.
        bases : tuple
            Tuple of base classes.

        Returns
        -------
        OrderedDict
            Ordered dictionary for the class namespace.
        """
        return OrderedDict()

    def __new__(cls: type, clsname: str, bases: tuple, clsdict: OrderedDict) -> Any:
        """
        Create a new class with descriptors and aggregators.

        Parameters
        ----------
        cls : type
            The metaclass itself.
        clsname : str
            Name of the class being created.
        bases : tuple
            Tuple of base classes.
        clsdict : OrderedDict
            Ordered dictionary of class attributes.

        Returns
        -------
        Any
            The newly created class object.
        """
        # Extract descriptor keys
        descriptors = [key for key, val in clsdict.items() if isinstance(val, Descriptor)]

        # Assign name attribute for each descriptor
        for name in descriptors:
            clsdict[name].name = name

        clsdict["_descriptors"] = descriptors

        # Extract aggregator keys
        aggregators = [key for key, val in clsdict.items() if isinstance(val, Aggregator)]

        # Assign name attribute for each descriptor
        for name in aggregators:
            clsdict[name].name = name

        clsdict["_aggregators"] = aggregators

        # Create the new class object
        clsobj = super().__new__(cls, clsname, bases, dict(clsdict))

        # Aggregate parameters from the current class and its bases
        parameters = []
        try:
            parameters += clsobj._parameters
        except AttributeError:
            pass

        for base in bases:
            base_parameters = getattr(base, "_parameters", [])
            parameters += base_parameters

        setattr(clsobj, "_parameters", parameters)

        # Categorize parameters based on their attributes
        sized_parameters = []
        polynomial_parameters = []
        required_parameters = []
        optional_parameters = []

        for param in parameters:
            descriptor = getattr(clsobj, param)

            # Skip if it's not an instance of Descriptor
            if not isinstance(descriptor, Descriptor):
                continue

            if hasattr(descriptor, "size"):
                sized_parameters.append(param)
            if hasattr(descriptor, "fill_values"):
                polynomial_parameters.append(param)
            if descriptor.default is None and not descriptor.is_optional:
                required_parameters.append(param)
            if descriptor.is_optional:
                optional_parameters.append(param)

        setattr(clsobj, "_sized_parameters", sized_parameters)
        setattr(clsobj, "_polynomial_parameters", polynomial_parameters)
        setattr(clsobj, "_required_parameters", required_parameters)
        setattr(clsobj, "_optional_parameters", optional_parameters)

        # Collect descriptors from base classes
        for base in bases:
            descriptors += getattr(base, "_descriptors", [])

        # Remove duplicates from the descriptors list
        args = list(dict.fromkeys(descriptors))

        # Register descriptor fields as arguments in __init__
        sig = make_signature(args)
        setattr(clsobj, "__signature__", sig)

        return clsobj


class AbstractStructMeta(StructMeta, ABCMeta):
    """Base class to allow for abstract metaclass structures."""

    pass


# %% Stucture / ParameterHandler
class Structure(metaclass=AbstractStructMeta):
    """
    A class representing a structured data entity.

    This class is designed to work in conjunction with the `StructMeta` metaclass
    to handle descriptors and related parameters.

    Attributes
    ----------
    _parameters : dict
        Dictionary of parameters associated with the instance.
    _sized_parameters : list
        List of parameters that have a `size` attribute.
    _polynomial_parameters : list
        List of parameters with `fill_values` attribute.
    _required_parameters : list
        List of parameters that have a default value of None.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """
        Initialize a Structure instance.

        Parameters are bound to the instance based on the `__signature__`
        defined by the metaclass.

        Parameters
        ----------
        *args
            Positional arguments representing parameters.
        **kwargs
            Keyword arguments representing parameters.
        """
        bound = self.__signature__.bind_partial(*args, **kwargs)
        for name, val in bound.arguments.items():
            setattr(self, name, val)

        self._parameters_dict = Dict()
        for param in self._parameters:
            value = getattr(self, param)
            if param in self._optional_parameters and value is None:
                continue
            self._parameters_dict[param] = value

    @property
    def parameters(self) -> dict:
        """dict: Parameters of the instance."""
        parameters = self._parameters_dict

        parameters.update(self.aggregated_parameters)

        return parameters

    @parameters.setter
    def parameters(self, parameters: dict) -> None:
        """
        Set parameters for the instance.

        Parameters
        ----------
        parameters : dict
            A dictionary of parameters to set.

        Raises
        ------
        ValueError
            If any of the provided parameters is not valid.
        """
        for param, value in parameters.items():
            if param not in self._parameters:
                raise ValueError("Not a valid parameter.")
            if value is not None:
                setattr(self, param, value)
                self._parameters_dict[param] = value

    @property
    def sized_parameters(self) -> Dict:
        """dict: Sized parameters of the instance."""
        parameters = {
            key: value for key, value in self.parameters.items() if key in self._sized_parameters
        }
        return Dict(parameters)

    @property
    def aggregated_parameters(self) -> Dict:
        """dict: Aggregated parameters of the instance."""
        parameters = {key: getattr(self, key) for key in self._aggregators}
        return Dict(parameters)

    @property
    def polynomial_parameters(self) -> Dict:
        """dict: Polynomial parameters of the instance."""
        parameters = {
            key: value
            for key, value in self.parameters.items()
            if key in self._polynomial_parameters
        }
        return Dict(parameters)

    @property
    def required_parameters(self) -> list:
        """list: Parameters that have no default value."""
        return self._required_parameters

    @property
    def missing_parameters(self) -> list:
        """list: Parameters that are required but not set."""
        missing_parameters = []
        for param in self.required_parameters:
            if getattr(self, param) is None:
                missing_parameters.append(param)

        return missing_parameters

    def check_required_parameters(self) -> bool:
        """
        Verify if all required parameters are set.

        Returns
        -------
        bool
            True if all required parameters are set. False otherwise.

        Warning:
            If any of the required parameters are missing.
        """
        if len(self.missing_parameters) == 0:
            return True
        else:
            for param in self.missing_parameters:
                warn(f'Missing parameter "{param}".')
            return False


def frozen_attributes(cls: type) -> type:
    """
    Class decorator to prevent setting new attributes after initialization.

    Parameters
    ----------
    cls : type
        The class to decorate.

    Returns
    -------
    type
        The decorated class with frozen attributes after initialization.
    """
    cls._is_frozen = False

    def frozensetattr(self: Structure, key: str, value: Any) -> None:
        if self._is_frozen and not hasattr(self, key):
            raise AttributeError(f"{cls.__name__} object has no attribute {key}")
        object.__setattr__(self, key, value)

    def init_decorator(func: Callable) -> Callable:
        @wraps(func)
        def wrapper(self: Structure, *args: Any, **kwargs: Any) -> None:
            func(self, *args, **kwargs)
            self._is_frozen = True

        return wrapper

    cls.__setattr__ = frozensetattr
    cls.__init__ = init_decorator(cls.__init__)

    return cls



"""

TODO:
- [ ] Add documentation / scope of tests (e.g. link to scipy/pymoo)

"""

import warnings
from functools import partial
from typing import NoReturn

import numpy as np
from CADETProcess.optimization import OptimizationProblem, OptimizationResults

__all__ = [
    "Rosenbrock",
    "LinearConstraintsSooTestProblem",
    "LinearConstraintsSooTestProblem2",
    "LinearEqualityConstraintsSooTestProblem",
    "NonlinearConstraintsSooTestProblem",
    "NonlinearLinearConstraintsSooTestProblem",
    "LinearConstraintsMooTestProblem",
    "LinearNonlinearConstraintsMooTestProblem",
    "NonlinearConstraintsMooTestProblem",
]


error = "Optimizer did not approach solution close enough."
default_test_kwargs = {"rtol": 0.01, "atol": 0.0001, "err_msg": error}


def allow_test_failure_percentage(test_function, test_kwargs, mismatch_tol=0.0):
    """When two arrays are compared, allow a certain fraction of the comparisons
    to fail. This behaviour is explicitly accepted in convergence tests of
    multi-objective tests. The reason behind it is that building the pareto
    front takes time and removing dominated solutions from the frontier can
    take a long time. Hence accepting a certain fraction of dominated-solutions
    is acceptable, when the majority points lies on the pareto front.

    While of course for full convergence checks the mismatch tolerance should
    be reduced to zero, for normal testing the fraction can be raised to
    say 0.25. This value can be adapted for easy or difficult cases.
    """
    assert 0.0 <= mismatch_tol <= 1.0, "mismatch_tol must be between 0 and 1."
    try:
        test_function(**test_kwargs)
    except AssertionError as e:
        msg = e.args[0].split("\n")
        lnum, mismatch_line = [
            (i, l) for i, l in enumerate(msg) if "Mismatched elements:" in l  # noqa: E741
        ][0]
        mismatch_percent = float(mismatch_line.split("(")[1].split("%")[0])
        if mismatch_percent / 100 > mismatch_tol:
            err_line = (
                "---> "
                + mismatch_line
                + f" exceeded tolerance ({mismatch_percent}% > {mismatch_tol * 100}%)"
            )
            msg[lnum] = err_line
            raise AssertionError("\n".join(msg))
        else:
            warn_line = (
                mismatch_line
                + f" below tolerance ({mismatch_percent}% <= {mismatch_tol * 100}%)"
            )
            warnings.warn(f"Equality test passed with {warn_line}")


class TestProblem(OptimizationProblem):
    # To prevent Pytest interpreting this class as test:
    __test__ = False

    @property
    def optimal_solution(self):
        """Must return X, F, and if it has, G."""
        raise NotImplementedError

    def test_if_solved(self, results):
        raise NotImplementedError

    @property
    def x0(self):
        raise NotImplementedError


class Rosenbrock(TestProblem):
    def __init__(self, *args, n_var=2, **kwargs):
        super().__init__("rosenbrock", *args, **kwargs)

        if n_var not in [1, 2, 3, 4, 5, 6, 7]:
            raise ValueError("n_var must be 1 or 2")

        self.add_variable("var_0", lb=-10, ub=10)
        if n_var == 2:
            self.add_variable("var_1", lb=-10, ub=10)

        self.add_objective(self._objective_function)

    def _objective_function(self, x):
        if self.n_variables == 1:
            return self.rosen_1D(x)
        if self.n_variables > 1 and self.n_variables <= 7:
            return self.rosen(x)

    @staticmethod
    def rosen(x):
        """Rosenbrock function for 2-D, 3-D and 4-7-D optimization problems.

        for 2-D and 3-D it has exactly 1 optimal solution which is
        f(x=(1,..,1)) = 0

        for 4-7 D it has the same global minimum but in addition a local optimum
        at x=(-1, 1, ..., 1)
        """
        return sum(100.0 * (x[1:] - x[:-1] ** 2.0) ** 2.0 + (1 - x[:-1]) ** 2.0)

    @classmethod
    def rosen_1D(cls, x):
        """Wrapper to ensure x is 2D before passing to actual function."""

        x_ = np.array([x, 1])
        return cls.rosen(x_)

    @property
    def optimal_solution(self):
        x = np.repeat(1, self.n_variables).reshape(1, self.n_variables)
        f = 0

        return x, f

    @property
    def x0(self):
        return np.repeat(0.9, self.n_variables)

    def test_if_solved(
        self, optimization_results: OptimizationResults, test_kwargs=default_test_kwargs
    ) -> NoReturn:
        x_true, f_true = self.optimal_solution
        x = optimization_results.x
        f = optimization_results.f

        test_kwargs["err_msg"] = error
        np.testing.assert_allclose(f, f_true, **test_kwargs)
        np.testing.assert_allclose(x, x_true, **test_kwargs)


class LinearConstraintsSooTestProblem(TestProblem):
    def __init__(
        self,
        transform=None,
        has_evaluator=False,
        significant_digits=None,
        *args,
        **kwargs,
    ):
        self.test_abs_tol = 0.1
        super().__init__("linear_constraints_single_objective", *args, **kwargs)
        self.setup_variables(transform=transform, significant_digits=significant_digits)
        self.setup_linear_constraints()
        if has_evaluator:
            eval_fun = lambda x: x
            self.add_evaluator(eval_fun)
            self.add_objective(self._objective_function, requires=eval_fun)
        else:
            self.add_objective(self._objective_function)

    def setup_variables(self, transform, significant_digits=None):
        self.add_variable(
            "var_0",
            lb=-2,
            ub=2,
            transform=transform,
            significant_digits=significant_digits,
        )
        self.add_variable(
            "var_1",
            lb=-2,
            ub=2,
            transform=transform,
            significant_digits=significant_digits,
        )
        self.add_variable(
            "var_2",
            lb=0,
            ub=2,
            transform="log",
            significant_digits=significant_digits,
        )

    def setup_linear_constraints(self):
        self.add_linear_constraint(["var_0", "var_1"], [-1, -0.5], 0)

    def _objective_function(self, x):
        return x[0] - x[1] + x[2]

    @property
    def optimal_solution(self):
        x = np.array([-1, 2, 0.0]).reshape(1, self.n_variables)
        f = -3

        return x, f

    @property
    def x0(self):
        return [-0.5, 1.5, 0.1]

    @property
    def conditional_minima(self):
        f_x0 = lambda x0: x0 - 2
        f_x1 = lambda x1: x1 * -3 / 2
        f_x2 = lambda x2: x2
        return f_x0, f_x1, f_x2

    def test_if_solved(
        self, optimization_results: OptimizationResults, test_kwargs=default_test_kwargs
    ):
        x_true, f_true = self.optimal_solution
        x = optimization_results.x
        f = optimization_results.f

        test_kwargs["err_msg"] = error
        np.testing.assert_allclose(f, f_true, **test_kwargs)
        np.testing.assert_allclose(x, x_true, **test_kwargs)


class NonlinearConstraintsSooTestProblem(TestProblem):
    def __init__(self, transform=None, has_evaluator=False, *args, **kwargs):
        self.fixture_evaluator = None
        super().__init__("linear_constraints_single_objective", *args, **kwargs)
        self.setup_variables(transform=transform)
        self.setup_evaluator(has_evaluator=has_evaluator)
        self.setup_nonlinear_constraints()
        self.setup_objectives()

    def setup_evaluator(self, has_evaluator):
        if has_evaluator:
            self.fixture_evaluator = lambda x: x
            self.add_evaluator(self.fixture_evaluator)
        else:
            self.fixture_evaluator = None

    def setup_objectives(self):
        self.add_objective(self._objective_function, requires=self.fixture_evaluator)

    def setup_variables(self, transform):
        self.add_variable("var_0", lb=-2, ub=0, transform=transform)
        self.add_variable("var_1", lb=-2, ub=2, transform=transform)

    def setup_nonlinear_constraints(self):
        """
        these should reproduce the same results as above only with nonlinear
        constraints.
        TODO: Bounds are probably redundant
        """
        nlc_fun_0 = lambda x: -1 * x[0] - 0.5 * x[1]
        self.add_nonlinear_constraint(
            nlc_fun_0,
            bounds=0,
            n_nonlinear_constraints=1,
            requires=self.fixture_evaluator,
        )

        def nlc_fun_1(x):
            return -0.01 / (1 + np.exp(x[0])) + 0.005, x[1]

        self.add_nonlinear_constraint(
            nlc_fun_1,
            bounds=[0.001, 2],
            n_nonlinear_constraints=2,
            requires=self.fixture_evaluator,
        )

    @property
    def x0(self):
        return [-0.5, 1.5]

    def _objective_function(self, x):
        return x[0] - x[1]

    @property
    def optimal_solution(self):
        x = np.array([-1, 2]).reshape(1, self.n_variables)
        f = -3

        return x, f

    def test_if_solved(
        self, optimization_results: OptimizationResults, test_kwargs=default_test_kwargs
    ):
        x_true, f_true = self.optimal_solution
        x = optimization_results.x
        f = optimization_results.f

        test_kwargs["err_msg"] = error
        np.testing.assert_allclose(f, f_true, **test_kwargs)
        np.testing.assert_allclose(x, x_true, **test_kwargs)


class LinearConstraintsSooTestProblem2(TestProblem):
    def __init__(self, transform=None, *args, **kwargs):
        super().__init__("linear_constraints_single_objective_2", *args, **kwargs)

        self.setup_variables(transform=transform)
        self.setup_linear_constraints()
        self.add_objective(self._objective_function)

    def setup_variables(self: OptimizationProblem, transform=None):
        self.add_variable("var_0", lb=-5, ub=5, transform=transform)
        self.add_variable("var_1", lb=-5, ub=5, transform=transform)
        self.add_variable("var_2", lb=-5, ub=5, transform=transform)

    def setup_linear_constraints(self):
        # cuts off upper right corner of var 0 and var 1
        self.add_linear_constraint(["var_0", "var_1"], [1, 2], 8)
        # halfs the cube along the diagonal plane and allows only values above
        self.add_linear_constraint(["var_0", "var_1", "var_2"], [-1, -1, -0.5], 0)
        # ???
        self.add_linear_constraint(["var_1", "var_2"], [0.5, -2], 4)

    def _objective_function(self, x):
        return 2 * x[0] - x[1] + 0.5 * x[2]

    @property
    def x0(self):
        return [-4, 4, 1]

    @property
    def optimal_solution(self):
        x = np.array([-5.0, 5.0, 0.0]).reshape(1, self.n_variables)
        f = -15.0

        return x, f

    def test_if_solved(
        self, optimization_results: OptimizationResults, test_kwargs=default_test_kwargs
    ):
        x_true, f_true = self.optimal_solution
        x = optimization_results.x
        f = optimization_results.f

        test_kwargs["err_msg"] = error
        np.testing.assert_allclose(f, f_true, **test_kwargs)
        np.testing.assert_allclose(x, x_true, **test_kwargs)


class LinearEqualityConstraintsSooTestProblem(TestProblem):
    def __init__(self, transform=None, *args, **kwargs):
        super().__init__(
            "linear_equality_constraints_single_objective", *args, **kwargs
        )

        self.setup_variables(transform=transform)
        self.setup_linear_constraints()
        self.add_objective(self._objective_function)

    def setup_variables(
        self: OptimizationProblem, transform=None, significant_digits=None
    ):
        self.add_variable(
            "var_0",
            lb=-5,
            ub=5,
            transform=transform,
            significant_digits=significant_digits,
        )
        self.add_variable(
            "var_1",
            lb=-5,
            ub=5,
            transform=transform,
            significant_digits=significant_digits,
        )
        self.add_variable(
            "var_2",
            lb=-5,
            ub=5,
            transform=transform,
            significant_digits=significant_digits,
        )

    def setup_linear_constraints(self):
        self.add_linear_equality_constraint(["var_0", "var_1"], [1.0, 2.0], 8, eps=1e-3)

    @property
    def x0(self):
        return np.array([-1.0, 4.5, -4.0])

    def _objective_function(self, x):
        return 2 * x[0] - x[1] + 0.5 * x[2]

    @property
    def optimal_solution(self):
        x = np.array([-2, 5, -5])
        f = self._objective_function(x)

        return x.reshape(1, self.n_variables), f

    def test_if_solved(
        self, optimization_results: OptimizationResults, test_kwargs=default_test_kwargs
    ):
        x_true, f_true = self.optimal_solution
        x = optimization_results.x
        f = optimization_results.f

        test_kwargs["err_msg"] = error
        np.testing.assert_allclose(f, f_true, **test_kwargs)
        np.testing.assert_allclose(x, x_true, **test_kwargs)


class NonlinearLinearConstraintsSooTestProblem(TestProblem):
    def __init__(self, transform=None, *args, **kwargs):
        self.test_tol = 0.1
        super().__init__(
            "nonlinear_linear_constraints_single_objective", *args, **kwargs
        )
        self.setup_variables(transform=transform)
        self.setup_linear_constraints()
        self.setup_nonlinear_constraints()
        self.add_objective(self._objective_function)

    def setup_variables(self, transform):
        self.add_variable("var_0", lb=-2, ub=2, transform=transform)
        self.add_variable("var_1", lb=-2, ub=2, transform=transform)

    def setup_linear_constraints(self):
        self.add_linear_constraint(["var_0", "var_1"], [-1, -0.5], 0)

    def setup_nonlinear_constraints(self):
        f_nonlinconc = lambda x: np.array([(x[0] + x[1]) ** 2])
        self.add_nonlinear_constraint(f_nonlinconc, "nonlincon_0", bounds=4)

    def _objective_function(self, x):
        return x[0] - x[1]

    @property
    def x0(self):
        return [-0.5, 1.5]

    @property
    def optimal_solution(self):
        x = np.array([-1, 2]).reshape(1, self.n_variables)
        f = -3

        return x, f

    def test_if_solved(
        self, optimization_results: OptimizationResults, test_kwargs=default_test_kwargs
    ):
        x_true, f_true = self.optimal_solution
        x = optimization_results.x
        f = optimization_results.f

        test_kwargs["err_msg"] = error
        np.testing.assert_allclose(f, f_true, **test_kwargs)
        np.testing.assert_allclose(x, x_true, **test_kwargs)


class LinearConstraintsMooTestProblem(TestProblem):
    """Function curtesy of Florian Schunck and Samuel Leweke."""

    def __init__(self, transform=None, *args, **kwargs):
        self.test_abs_tol = 0.1

        super().__init__("linear_constraints_multi_objective", *args, **kwargs)
        self.setup_variables(transform=transform)
        self.setup_linear_constraints()
        self.setup_objectives()

    def setup_variables(self: OptimizationProblem, transform=None):
        self.add_variable("var_0", lb=1, ub=5, transform=transform)
        self.add_variable("var_1", lb=0, ub=3, transform=transform)

    def setup_linear_constraints(self):
        self.add_linear_constraint(["var_0", "var_1"], [-1, -1], -3)
        self.add_linear_constraint(["var_0", "var_1"], [1, -1], 5)

    @staticmethod
    def _objective_function(x):
        f1 = x[0]
        f2 = (1 + x[1]) / x[0]

        return f1, f2

    def setup_objectives(self):
        def f1(x):
            return self._objective_function(x)[0]

        def f2(x):
            return self._objective_function(x)[1]

        self.add_objective(f1, n_objectives=1)
        self.add_objective(f2, n_objectives=1)

    def find_corresponding_x2(self, x1):
        """
        in a point x in a pareto set
        """
        return np.where(x1 <= 3, 3 - x1, 0)

    @property
    def conditional_minima(self):
        def f_x0(x0):
            f1 = x0
            # solve constraints with respect to x1 and substitute in a way
            # that minimizes f2
            # when x0 <= 3 the first linear constraint is dominating,
            # when x0 > 3 the boundary constraint of x1 is dominating
            f2 = np.where(x0 <= 3, (1 + -x0 + 3) / x0, (1 + 0) / x0)
            return np.array([f1, f2])

        def f_x1(x1):
            f1 = np.where(x1 <= 2, -x1 + 3, 1)
            f2 = (1 + x1) / 5
            return np.array([f1, f2])

        return f_x0, f_x1

    @property
    def x0(self):
        return [1.6, 1.4]

    @property
    def optimal_solution(self):
        x1 = np.linspace(1, 5, 101)
        x2 = self.find_corresponding_x2(x1=x1)
        X = np.column_stack([x1, x2])

        F = np.array(list(map(self._objective_function, X)))

        return X, F

    def test_if_solved(
        self, optimization_results: OptimizationResults, test_kwargs=default_test_kwargs
    ) -> NoReturn:
        X = optimization_results.x

        x1, x2 = X.T
        x2_test = np.where(x1 <= 3, 3 - x1, 0)

        test_kwargs_ = test_kwargs.copy()
        test_kwargs_["err_msg"] = error
        mismatch_tol = test_kwargs_.pop("mismatch_tol", 0.0)
        test_func = partial(np.testing.assert_allclose, actual=x2, desired=x2_test)

        allow_test_failure_percentage(
            test_function=test_func, test_kwargs=test_kwargs_, mismatch_tol=mismatch_tol
        )


class LinearNonlinearConstraintsMooTestProblem(TestProblem):
    """Function curtesy of Florian Schunck and Samuel Leweke."""

    def __init__(self, has_evaluator=False, *args, **kwargs):
        super().__init__("linear_constraints_multi_objective", *args, **kwargs)
        self.setup_variables()
        self.setup_linear_constraints()
        self.setup_nonlinear_constraints()
        self.setup_objectives(has_evaluator=has_evaluator)

    def setup_variables(self):
        self.add_variable("var_0", lb=1, ub=5)
        self.add_variable("var_1", lb=0, ub=3)

    def setup_linear_constraints(self):
        self.add_linear_constraint(["var_0", "var_1"], [-1, -1], -2)
        self.add_linear_constraint(["var_0", "var_1"], [1, -1], 5)

    def setup_nonlinear_constraints(self):
        f_nonlinconc_0 = lambda x: np.array([x[0] ** 2, x[1] ** 2])
        f_nonlinconc_1 = lambda x: np.array([x[0] ** 1.1, x[1] ** 1.1])

        self.add_nonlinear_constraint(
            nonlincon=f_nonlinconc_0,
            name="nonlincon_0",
            bounds=4,
            n_nonlinear_constraints=2,
        )

        self.add_nonlinear_constraint(
            nonlincon=f_nonlinconc_1,
            name="nonlincon_1",
            bounds=3,
            n_nonlinear_constraints=2,
        )

    def setup_objectives(self, has_evaluator):
        if has_evaluator:
            eval_fun = lambda x: x
            self.add_evaluator(eval_fun)
            self.add_objective(
                objective=self._objective_function,
                requires=eval_fun,
                n_objectives=2,
            )
        else:
            self.add_objective(self._objective_function, n_objectives=2)

    @staticmethod
    def _objective_function(x):
        f1 = x[0]
        f2 = (1 + x[1]) / x[0]

        return f1, f2

    def find_corresponding_x2(self, x1):
        """
        in a point x in a pareto set
        """
        return np.where(x1 <= 2, 2 - x1, 0)

    @property
    def x0(self):
        return [1.6, 1.4]

    @property
    def optimal_solution(self):
        x1 = np.linspace(1, 5, 101)
        x2 = self.find_corresponding_x2(x1=x1)
        X = np.column_stack([x1, x2])

        F = np.array(list(map(self._objective_function, X)))

        return X, F

    def test_if_solved(
        self, optimization_results: OptimizationResults, test_kwargs=default_test_kwargs
    ):
        X = optimization_results.x

        x1, x2 = X.T
        x2_test = self.find_corresponding_x2(x1)

        test_kwargs_ = test_kwargs.copy()
        test_kwargs_["err_msg"] = error
        mismatch_tol = test_kwargs_.pop("mismatch_tol", 0.0)
        test_func = partial(np.testing.assert_allclose, actual=x2, desired=x2_test)

        allow_test_failure_percentage(
            test_function=test_func, test_kwargs=test_kwargs_, mismatch_tol=mismatch_tol
        )


class NonlinearConstraintsMooTestProblem(TestProblem):
    def __init__(self, has_evaluator=False, *args, **kwargs):
        from pymoo.problems.multi import SRN

        self._problem = SRN()
        self.fixture_evaluator = None
        super().__init__("nonlinear_constraints_multi_objective", *args, **kwargs)

        self.add_variable("var_0", lb=-20, ub=20)
        self.add_variable("var_1", lb=-20, ub=20)
        self.setup_evaluator(has_evaluator=has_evaluator)
        self.setup_nonlinear_constraints()
        self.setup_objectives()

    def setup_evaluator(self, has_evaluator):
        if has_evaluator:
            self.fixture_evaluator = lambda x: x
            self.add_evaluator(self.fixture_evaluator)
        else:
            self.fixture_evaluator = None

    def setup_nonlinear_constraints(self):
        self.add_nonlinear_constraint(
            nonlincon=self._nonlincon_fun,
            requires=self.fixture_evaluator,
            n_nonlinear_constraints=2,
        )

    def setup_objectives(self):
        self.add_objective(
            objective=self._objective_function,
            requires=self.fixture_evaluator,
            n_objectives=2,
        )

    def _objective_function(self, x):
        return self._problem.evaluate(x)[0]

    def _nonlincon_fun(self, x):
        return self._problem.evaluate(x)[1]

    @property
    def x0(self):
        return [-2.4, 5.0]

    @property
    def optimal_solution(self):
        X = self._problem.pareto_set()
        F = self._problem.pareto_front()
        # TODO: test nonlinear constraints as well.

        return X, F  # G ???

    def test_if_solved(
        self, optimization_results: OptimizationResults, test_kwargs=default_test_kwargs
    ) -> NoReturn:
        X = optimization_results.x_transformed
        x1, x2 = X.T

        test_kwargs_ = test_kwargs.copy()
        mismatch_tol = test_kwargs_.pop("mismatch_tol", 0.0)
        test_kwargs_["err_msg"] = error

        test_func_1 = partial(np.testing.assert_allclose, actual=x1, desired=-2.5)
        test_func_2 = partial(np.testing.assert_array_less, x=x2, y=14.7902)
        test_func_3 = partial(np.testing.assert_array_less, x=-x2, y=-2.5)

        allow_test_failure_percentage(
            test_function=test_func_1,
            test_kwargs=test_kwargs_,
            mismatch_tol=mismatch_tol,
        )

        allow_test_failure_percentage(
            test_function=test_func_2, test_kwargs={}, mismatch_tol=mismatch_tol
        )

        allow_test_failure_percentage(
            test_function=test_func_3, test_kwargs={}, mismatch_tol=mismatch_tol
        )



from abc import abstractmethod
from typing import Any, Optional

import numpy as np

from CADETProcess import CADETProcessError, SimulationResults
from CADETProcess.dataStructure import Bool, Structure, UnsignedFloat, UnsignedInteger
from CADETProcess.log import get_logger, log_exceptions, log_results, log_time
from CADETProcess.processModel import Process
from CADETProcess.stationarity import NRMSE, RelativeArea, StationarityEvaluator


class SimulatorBase(Structure):
    """
    Base class for Solver APIs.

    Holds the configuration of the individual solvers and provides an interface
    for calling the run method. The class converts the process configuration
    into the API's configuration format and converts the results back to the
    CADETProcess format.

    Attributes
    ----------
    time_resolution : float
        Time interval for user solution times. The default value is 1 second.
    resolution_cutoff : float
        Solution times closer to section times than the cutoff value are removed
        to avoid IDAS errors. The default value is 1e-3 seconds.
    n_cycles : int
        Number of cycles to be simulated. The default is 1.
    evaluate_stationarity : bool
        If True, simulate until stationarity is reached.
        The default is False
    n_cycles_min : int
        Minimum number of cycles to be simulated if evaluate_stationarity is True.
        The default is 5.
    n_cycles_max : int
        Maximum number of cycles to be simulated if evaluate_stationarity is True.
        The default is 100.
    raise_exception_on_max_cycles : bool
        Raise an exception when the maximum number of cycles is exceeded.
        The default is False

    See Also
    --------
    CADETProcess.processModel.Process
    CADETProcess.stationarity.StationarityEvaluator
    """

    time_resolution = UnsignedFloat(default=1)
    resolution_cutoff = UnsignedFloat(default=1e-3)

    n_cycles = UnsignedInteger(default=1)
    evaluate_stationarity = Bool(default=False)
    n_cycles_min = UnsignedInteger(default=1)
    n_cycles_batch = UnsignedInteger(default=5)
    n_cycles_max = UnsignedInteger(default=100)
    raise_exception_on_max_cycles = Bool(default=False)

    def __init__(
        self,
        stationarity_evaluator: Optional[StationarityEvaluator] = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """Initialize Simulator Base."""
        super().__init__(*args, **kwargs)

        self.logger = get_logger("Simulation")

        if stationarity_evaluator is None:
            self.stationarity_evaluator = StationarityEvaluator()
            self.stationarity_evaluator.add_criterion(RelativeArea())
            self.stationarity_evaluator.add_criterion(NRMSE())
        else:
            self.stationarity_evaluator = stationarity_evaluator
            self.evaluate_stationarity = True

    @property
    def sig_fig(self) -> int:
        """int: Number of significant figures based on resolution_cutoff."""
        return int(-np.log10(self.resolution_cutoff))

    def get_solution_time(self, process: Process, cycle: int = 1) -> np.ndarray:
        """
        Get the time vector for one cycle of a process.

        Parameters
        ----------
        process : Process
            The process to simulate.
        cycle : int, optional
            The cycle number, by default 1.

        Returns
        -------
        np.ndarray
            Time vector for one cycle.

        See Also
        --------
        CADETProcess.processModel.Process.section_times
        get_solution_time_complete
        """
        solution_times = np.arange(
            (cycle - 1) * process.cycle_time,
            cycle * process.cycle_time,
            self.time_resolution,
        )
        section_times = self.get_section_times(process)
        solution_times = np.append(solution_times, section_times)
        solution_times = np.round(solution_times, self.sig_fig)
        solution_times = np.sort(solution_times)
        solution_times = np.unique(solution_times)

        diff = np.where(np.diff(solution_times) < self.resolution_cutoff)[0]
        indices = []
        for d in diff:
            if solution_times[d] in process.section_times:
                indices.append(d + 1)
            else:
                indices.append(d)

        solution_times = np.delete(solution_times, indices)

        return solution_times

    def get_solution_time_complete(self, process: Process) -> np.ndarray:
        """
        Get the time vector for multiple cycles of a process.

        Parameters
        ----------
        process : Process
            The process to simulate.

        Returns
        -------
        np.ndarray
            Time vector for multiple cycles of a process.

        See Also
        --------
        n_cycles
        get_section_times_complete
        get_solution_time
        """
        time = self.get_solution_time(process)

        solution_times = time

        for i in range(1, self.n_cycles):
            solution_times = np.append(solution_times, (i) * time[-1] + time[1:])
        solution_times = np.round(solution_times, self.sig_fig)

        return solution_times.tolist()

    def get_section_times(self, process: Process) -> list:
        """
        Get the section times for a single cycle of a process.

        Parameters
        ----------
        process : Process
            The process to simulate.

        Returns
        -------
        list
            Section times for a single cycle of a process.

        See Also
        --------
        get_section_times_complete
        get_solution_time_complete
        CADETProcess.processModel.Process.section_times
        """
        section_times = np.array(process.section_times)
        section_times = np.round(section_times, self.sig_fig)

        return section_times.tolist()

    def get_section_times_complete(self, process: Process) -> list:
        """
        Get the section times for multiple cycles of a process.

        Parameters
        ----------
        process : Process
            The process to simulate.

        Returns
        -------
        list
            Section times for multiple cycles of a process.

        See Also
        --------
        get_section_times
        n_cycles
        get_solution_time_complete
        CADETProcess.processModel.Process.section_times
        """
        sections = np.array(self.get_section_times(process))
        cycle_time = sections[-1]

        section_times_complete = []
        for cycle in range(self.n_cycles):
            section_cycle = cycle * cycle_time + sections[0:-1]
            section_times_complete += section_cycle.tolist()

        section_times_complete.append(self.n_cycles * sections[-1])

        section_times_complete = np.round(section_times_complete, self.sig_fig)

        return section_times_complete.tolist()

    def simulate(
        self,
        process: Process,
        previous_results: Optional[SimulationResults] = None,
        **kwargs: Any,
    ) -> SimulationResults:
        """
        Simulate the process.

        Depending on the state of evaluate_stationarity, the process is
        simulated until the termination criterion is reached.

        Parameters
        ----------
        process : Process
            The process to be simulated.
        previous_results : SimulationResults, optional
            Results of the previous simulation run for initial conditions.
        **kwargs
            Additional keyword arguments.

        Returns
        -------
        results : SimulationResults
            Results of the final cycle of the simulation.

        Raises
        ------
        TypeError
            If the process is not an instance of Process.
        CADETProcessError
            If the process is not configured correctly.

        See Also
        --------
        simulate_n_cycles
        simulate_to_stationarity
        run
        """
        if not isinstance(process, Process):
            raise TypeError("Expected Process")

        process.lock = True
        if not process.check_config():
            raise CADETProcessError("Process is not configured correctly.")

        if self.n_cycles_max < self.n_cycles_min:
            raise ValueError(
                "n_cycles_max is set lower than n_cycles_min "
                f"({self.n_cycles_max} vs {self.n_cycles_min}). "
            )

        # If "max" is below "batch", reduce "batch" to "max" to only run "max" cycles.
        if self.n_cycles_max < self.n_cycles_batch:
            self.n_cycles_batch = self.n_cycles_max

        if not self.evaluate_stationarity:
            results = self.simulate_n_cycles(
                process, self.n_cycles, previous_results, **kwargs
            )
        else:
            results = self.simulate_to_stationarity(process, previous_results, **kwargs)
        process.lock = False

        return results

    @log_time("Simulation")
    @log_results("Simulation")
    @log_exceptions("Simulation")
    def simulate_n_cycles(
        self,
        process: Process,
        n_cyc: int,
        previous_results: Optional[SimulationResults] = None,
        **kwargs: Any,
    ) -> SimulationResults:
        """
        Simulate the process for a given number of cycles.

        Parameters
        ----------
        process : Process
            The process to be simulated.
        n_cyc : int
            The number of cycles to simulate.
        previous_results : SimulationResults, optional
            Results of the previous simulation run.
        **kwargs
            Additional keyword arguments.

        Returns
        -------
        results : SimulationResults
            Results of the final cycle of the simulation.

        Raises
        ------
        TypeError
            If the process is not an instance of Process.

        See Also
        --------
        simulate
        simulate_to_stationarity
        StationarityEvaluator
        run
        """
        n_cyc_orig = self.n_cycles
        self.n_cycles = n_cyc

        if not isinstance(process, Process):
            raise TypeError("Expected Process")

        if previous_results is not None:
            self.set_state_from_results(process, previous_results)

        return self._run(process, **kwargs)

        self.n_cycles = n_cyc_orig

    @log_time("Simulation")
    @log_results("Simulation")
    @log_exceptions("Simulation")
    def simulate_to_stationarity(
        self,
        process: Process,
        results: Optional[SimulationResults] = None,
        **kwargs: Any,
    ) -> SimulationResults:
        """
        Simulate the process until stationarity is reached.

        Parameters
        ----------
        process : Process
            The process to be simulated.
        results : SimulationResults, optional
            Results of the previous simulation run.
        **kwargs
            Additional keyword arguments.

        Returns
        -------
        results : SimulationResults
            Results of the final cycle of the simulation.

        Raises
        ------
        TypeError
            If the process is not an instance of Process.
        CADETProcessError
            If the simulation doesn't terminate successfully and
            `raise_exception_on_max_cycles` is True.

        See Also
        --------
        simulate
        run
        StationarityEvaluator
        """
        if not isinstance(process, Process):
            raise TypeError("Expected Process")

        n_cyc_orig = self.n_cycles
        self.n_cycles = max(self.n_cycles_min, self.n_cycles_batch)

        if results is not None:
            n_cyc = results.n_cycles
        else:
            n_cyc = 0

        while True:
            n_cyc += self.n_cycles_batch

            if results is not None:
                self.set_state_from_results(process, results)

            new_results = self._run(process, **kwargs)

            if results is None:
                results = new_results
            else:
                results.update(new_results)

            if n_cyc == 1:
                continue

            stationarity = self.stationarity_evaluator.assert_stationarity(results)

            if stationarity:
                break

            if n_cyc >= self.n_cycles_max:
                msg = "Exceeded maximum number of cycles."
                if self.raise_exception_on_max_cycles:
                    raise CADETProcessError(msg)
                else:
                    self.logger.warning(msg)
                break

        self.n_cycles = n_cyc_orig

        return results

    def set_state_from_results(
        self,
        process: Process,
        results: SimulationResults,
    ) -> Process:
        """
        Set the process state from the simulation results.

        Parameters
        ----------
        process : Process
            The process to set the state for.
        results : SimulationResults
            The simulation results containing the state information.

        Returns
        -------
        Process
            The process with the updated state.
        """
        process.system_state = results.system_state["state"]
        process.system_state_derivative = results.system_state["state_derivative"]

        return process

    @abstractmethod
    def _run(process: Process, **kwargs: Any) -> SimulationResults:
        """
        Abstract method for running a simulation.

        Parameters
        ----------
        process : Process
            The process to be simulated.
        **kwargs
            Additional keyword arguments.

        Returns
        -------
        SimulationResults
            The simulation results including process and solver configuration.

        Raises
        ------
        TypeError
            If the process is not an instance of Process.
        CADETProcessError
            If the simulation doesn't terminate successfully.
        """
        return

    evaluate = simulate
    __call__ = simulate

    @property
    def stationarity_evaluator(self) -> StationarityEvaluator:
        """StationarityEvaluator: The stationarity evaluator."""
        return self._stationarity_evaluator

    @stationarity_evaluator.setter
    def stationarity_evaluator(
        self,
        stationarity_evaluator: StationarityEvaluator,
    ) -> None:
        """
        Set the stationarity evaluator.

        Parameters
        ----------
        stationarity_evaluator : StationarityEvaluator
            The stationarity evaluator to set.

        Raises
        ------
        CADETProcessError
            If the stationarity evaluator is not an instance of StationarityEvaluator.
        """
        if not isinstance(stationarity_evaluator, StationarityEvaluator):
            raise CADETProcessError("Expected StationarityEvaluator")
        self._stationarity_evaluator = stationarity_evaluator



# ---
# jupyter:
#   jupytext:
#     formats: md:myst,py:percent
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.17.1
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# %% [markdown]
# # Fit Particle Porosity
#
# ## Experiment
# Experiment:
# - Acetone (pore penetrating tracer './experimental_data/pore_penetrating_tracer.csv')
# - data time / s und c / mM

# %%
import numpy as np
data = np.loadtxt('experimental_data/pore_penetrating_tracer.csv', delimiter=',')

time_experiment = data[:, 0]
c_experiment = data[:, 1]

from CADETProcess.reference import ReferenceIO
tracer_peak = ReferenceIO(
    'Tracer Peak', time_experiment, c_experiment
)

if __name__ == '__main__':
    _ = tracer_peak.plot()

# %% [markdown]
# ## Reference Model
#
# Here, initial values for `axial_dispersion` and `particle_porosity` are assumed.
# They will later be optimized.

# %%
from CADETProcess.processModel import ComponentSystem
component_system = ComponentSystem(['Penetrating Tracer'])

# %%
from CADETProcess.processModel import Inlet, Outlet, LumpedRateModelWithPores

feed = Inlet(component_system, name='feed')
feed.c = [131.75]

eluent = Inlet(component_system, name='eluent')
eluent.c = [0]

column = LumpedRateModelWithPores(component_system, name='column')

column.length = 0.1
column.diameter = 0.0077
column.particle_radius = 34e-6

column.axial_dispersion = 1e-8
column.bed_porosity = 0.3
column.particle_porosity = 0.8

column.film_diffusion = [1]

outlet = Outlet(component_system, name='outlet')

# %%
from CADETProcess.processModel import FlowSheet

flow_sheet = FlowSheet(component_system)

flow_sheet.add_unit(feed)
flow_sheet.add_unit(eluent)
flow_sheet.add_unit(column)
flow_sheet.add_unit(outlet)

flow_sheet.add_connection(feed, column)
flow_sheet.add_connection(eluent, column)
flow_sheet.add_connection(column, outlet)

# %%
from CADETProcess.processModel import Process

Q_ml_min = 0.5  # ml/min
Q_m3_s = Q_ml_min / (60 * 1e6)
V_tracer = 50e-9  # m3

process = Process(flow_sheet, 'Tracer')
process.cycle_time = 15 * 60

process.add_event(
    'feed_on',
    'flow_sheet.feed.flow_rate',
    Q_m3_s, 0
)
process.add_event(
    'feed_off',
    'flow_sheet.feed.flow_rate',
    0,
    V_tracer / Q_m3_s
)

process.add_event(
    'feed_water_on',
    'flow_sheet.eluent.flow_rate',
     Q_m3_s,
     V_tracer / Q_m3_s
)

process.add_event(
    'eluent_off',
    'flow_sheet.eluent.flow_rate',
    0,
    process.cycle_time
)

# %% [markdown]
# ## Simulator

# %%
from CADETProcess.simulator import Cadet
simulator = Cadet()

if __name__ == '__main__':
    simulation_results = simulator.simulate(process)
    _ = simulation_results.solution.outlet.inlet.plot()

# %% [markdown]
# ## Comparator

# %%
from CADETProcess.comparison import Comparator

comparator = Comparator()
comparator.add_reference(tracer_peak)
comparator.add_difference_metric(
    'NRMSE', tracer_peak, 'outlet.outlet',
)

if __name__ == '__main__':
    comparator.plot_comparison(simulation_results)

# %% [markdown]
# ## Optimization Problem

# %%
from CADETProcess.optimization import OptimizationProblem

optimization_problem = OptimizationProblem('particle_porosity')

optimization_problem.add_evaluation_object(process)

optimization_problem.add_variable(
    name='particle_porosity',
    parameter_path='flow_sheet.column.particle_porosity',
    lb=0.5, ub=0.99,
    transform='auto'
)

optimization_problem.add_evaluator(simulator)

optimization_problem.add_objective(
    comparator,
    n_objectives=comparator.n_metrics,
    requires=[simulator]
)


def callback(simulation_results, individual, evaluation_object, callbacks_dir='./'):
    comparator.plot_comparison(
        simulation_results,
        file_name=f'{callbacks_dir}/{individual.id}_{evaluation_object}_comparison.png',
        show=False
    )


optimization_problem.add_callback(callback, requires=[simulator])

# %% [markdown]
# ## Optimizer

# %%
from CADETProcess.optimization import U_NSGA3
optimizer = U_NSGA3()

# %% [markdown]
# ```{note}
# For performance reasons, the optimization is currently not run when building the documentation.
# In future, we will try to sideload pre-computed results to also discuss them here.
# ```
#
# ```
# if __name__ == '__main__':
#     optimization_results = optimizer.optimize(
#         optimization_problem,
#         use_checkpoint=True
#     )
# ```
